from __future__ import annotations

import datetime
import json
import math
import os
import random
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Dict, List, Tuple

import netsquid as ns

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NetworkScheduleConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchInfo, BatchResult, ProgramBatch, ProgramInput
from qoala.sim.build import build_network_from_config
from qoala.sim.network import ProcNodeNetwork
from qoala.util.logging import LogManager


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def topology_config(num_qubits: int) -> TopologyConfig:
    return TopologyConfig.perfect_config_uniform(
        num_qubits,
        single_instructions=[
            "INSTR_INIT",
            "INSTR_ROT_X",
            "INSTR_ROT_Y",
            "INSTR_ROT_Z",
            "INSTR_X",
            "INSTR_Y",
            "INSTR_Z",
            "INSTR_H",
            "INSTR_MEASURE",
        ],
        single_duration=1e3,
        two_instructions=["INSTR_CNOT", "INSTR_CZ"],
        two_duration=100e3,
    )


def get_client_config(id: int) -> ProcNodeConfig:
    # client only needs 1 qubit
    return ProcNodeConfig(
        node_name=f"client_{id}",
        node_id=id,
        topology=topology_config(20),
        latencies=LatenciesConfig(
            host_instr_time=500, host_peer_latency=30_000, qnos_instr_time=1000
        ),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def get_server_config(id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name="server",
        node_id=id,
        topology=topology_config(num_qubits),
        latencies=LatenciesConfig(
            host_instr_time=500, host_peer_latency=30_000, qnos_instr_time=1000
        ),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def create_network(
    server_cfg: ProcNodeConfig,
    client_configs: List[ProcNodeConfig],
    num_clients: int,
    num_iterations: List[int],
    cc: float,
    use_netschedule: bool,
    bin_length: float,
) -> ProcNodeNetwork:
    assert len(client_configs) == num_clients

    node_cfgs = [server_cfg] + client_configs
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=node_cfgs, link_duration=1000
    )

    pattern = []
    server_pid_index = 0
    for i in range(num_clients):
        for j in range(num_iterations[i]):
            pattern.append((i + 1, j, 0, server_pid_index))
            server_pid_index += 1

    if use_netschedule:
        network_cfg.netschedule = NetworkScheduleConfig(
            bin_length=bin_length,
            first_bin=0,
            bin_pattern=pattern,
            repeat_period=bin_length * num_clients * max(num_iterations),
        )

    cconns = [
        ClassicalConnectionConfig.from_nodes(i, 0, cc)
        for i in range(1, num_clients + 1)
    ]
    network_cfg.cconns = cconns
    return build_network_from_config(network_cfg)


@dataclass
class BqcResult:
    client_batches: List[Dict[int, ProgramBatch]]
    client_results: List[Dict[int, BatchResult]]


def load_server_program(remote_name: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), "vbqc_server.iqoala")
    with open(path) as file:
        server_text = file.read()
    program = QoalaParser(server_text).parse()

    # Replace "client" by e.g. "client_1"
    program.meta.csockets[0] = remote_name
    program.meta.epr_sockets[0] = remote_name

    return program


def load_client_program() -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), "vbqc_client.iqoala")
    with open(path) as file:
        client_text = file.read()
    return QoalaParser(client_text).parse()


def create_server_batch(
    client_id: int,
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
) -> BatchInfo:
    server_program = load_server_program(remote_name=f"client_{client_id}")
    return BatchInfo(
        program=server_program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=0,
    )


def create_client_batch(
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
) -> BatchInfo:
    client_program = load_client_program()
    return BatchInfo(
        program=client_program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=0,
    )


def run_bqc(
    alpha: int,
    beta: int,
    theta1: int,
    theta2: int,
    dummy0: int,
    dummy1: int,
    num_iterations: List[int],
    num_clients: int,
    linear: bool,
    cc: float,
    server_num_qubits: int,
    use_netschedule: bool,
    bin_length: float,
):
    ns.sim_reset()
    ns.set_qstate_formalism(ns.QFormalism.DM)
    seed = random.randint(0, 1000)
    ns.set_random_state(seed=seed)

    server_config = get_server_config(id=0, num_qubits=server_num_qubits)
    client_configs = [get_client_config(i) for i in range(1, num_clients + 1)]

    network = create_network(
        server_config,
        client_configs,
        num_clients,
        num_iterations,
        cc,
        use_netschedule,
        bin_length,
    )
    server_procnode = network.nodes["server"]
    server_batches: Dict[int, ProgramBatch] = {}  # client ID -> server batch
    client_batches: Dict[int, ProgramBatch] = {}  # client ID -> client batch

    for client_id in range(1, num_clients + 1):
        # index in num_iterations list
        index = client_id - 1

        server_inputs = [
            ProgramInput({"client_id": client_id}) for _ in range(num_iterations[index])
        ]

        server_unit_module = UnitModule.from_full_ehi(server_procnode.memmgr.get_ehi())
        server_batch_info = create_server_batch(
            client_id=client_id,
            inputs=server_inputs,
            unit_module=server_unit_module,
            num_iterations=num_iterations[index],
        )

        server_batches[client_id] = server_procnode.submit_batch(server_batch_info)

    for client_id in range(1, num_clients + 1):
        # index in num_iterations list
        index = client_id - 1

        client_inputs = [
            ProgramInput(
                {
                    "server_id": 0,
                    "alpha": alpha,
                    "beta": beta,
                    "theta1": theta1,
                    "theta2": theta2,
                    "dummy0": dummy0,
                    "dummy1": dummy1,
                }
            )
            for _ in range(num_iterations[index])
        ]

        client_procnode = network.nodes[f"client_{client_id}"]

        client_unit_module = UnitModule.from_full_ehi(client_procnode.memmgr.get_ehi())
        client_batch_info = create_client_batch(
            client_inputs, client_unit_module, num_iterations[index]
        )

        client_batches[client_id] = client_procnode.submit_batch(client_batch_info)
        batch_id = client_batches[client_id].batch_id
        server_pids = [inst.pid for inst in server_batches[client_id].instances]
        # print(
        #     f"client ID: {client_id}, batch ID: {batch_id}, server PIDs: {server_pids}"
        # )
        client_procnode.initialize_processes(
            remote_pids={batch_id: server_pids}, linear=linear
        )

    client_pids = {
        server_batches[client_id].batch_id: [
            inst.pid for inst in client_batches[client_id].instances
        ]
        for client_id in range(1, num_clients + 1)
    }
    # print(f"client PIDs: {client_pids}")
    server_procnode.initialize_processes(remote_pids=client_pids, linear=linear)

    network.start()
    start_time = ns.sim_time()
    ns.sim_run()
    end_time = ns.sim_time()
    makespan = end_time - start_time

    client_procnodes = [network.nodes[f"client_{i}"] for i in range(1, num_clients + 1)]
    client_batches = [node.get_batches() for node in client_procnodes]

    client_results: List[Dict[int, BatchResult]]
    client_results = [node.scheduler.get_batch_results() for node in client_procnodes]

    return BqcResult(client_batches, client_results), makespan


def check_computation(
    alpha: int,
    beta: int,
    theta1: int,
    theta2: int,
    dummy0: int,
    dummy1: int,
    expected: int,
    num_iterations: int,
    num_clients: int,
    linear: bool,
    cc: float,
    server_num_qubits: int,
    use_netschedule: bool,
    bin_length: float,
):
    ns.sim_reset()
    bqc_result, makespan = run_bqc(
        alpha=alpha,
        beta=beta,
        theta1=theta1,
        theta2=theta2,
        dummy0=dummy0,
        dummy1=dummy1,
        num_iterations=num_iterations,
        num_clients=num_clients,
        linear=linear,
        cc=cc,
        server_num_qubits=server_num_qubits,
        use_netschedule=use_netschedule,
        bin_length=bin_length,
    )

    batch_success_probabilities: List[float] = []

    for i in range(num_clients):
        assert len(bqc_result.client_results[i]) == 1
        batch_result = bqc_result.client_results[i][0]
        assert len(bqc_result.client_batches[i]) == 1
        program_batch = bqc_result.client_batches[i][0]
        batch_iterations = program_batch.info.num_iterations

        m2s = [result.values["m2"] for result in batch_result.results]
        correct_outcomes = len([m2 for m2 in m2s if m2 == expected])
        succ_prob = round(correct_outcomes / batch_iterations, 2)
        batch_success_probabilities.append(succ_prob)

    return batch_success_probabilities, makespan


def compute_succ_prob_computation(
    num_clients: int,
    num_iterations: List[int],
    linear: bool,
    cc: float,
    server_num_qubits: int,
    use_netschedule: bool,
    bin_length: float,
) -> Tuple[float, float]:
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    all_probs = []
    all_makespans = []
    probs, makespan = check_computation(
        alpha=8,
        beta=24,
        theta1=2,
        theta2=22,
        dummy0=0,
        dummy1=0,
        expected=1,
        num_iterations=num_iterations,
        num_clients=num_clients,
        linear=linear,
        cc=cc,
        server_num_qubits=server_num_qubits,
        use_netschedule=use_netschedule,
        bin_length=bin_length,
    )
    all_probs.append(probs[0])
    all_makespans.append(makespan)

    prob = sum(all_probs) / len(all_probs)
    makespan = sum(all_makespans)
    return prob, makespan


def bqc_computation(
    num_clients: int,
    num_iterations: int,
    linear: bool,
    cc: float,
    server_num_qubits: int,
    use_netschedule: bool,
    bin_length: float,
) -> float:
    succ_probs, makespan = compute_succ_prob_computation(
        num_clients=num_clients,
        num_iterations=[num_iterations] * num_clients,
        linear=linear,
        cc=cc,
        server_num_qubits=server_num_qubits,
        use_netschedule=use_netschedule,
        bin_length=bin_length,
    )
    # print(f"success probabilities: {succ_probs}")
    # print(f"makespan: {makespan:_}")
    return makespan


def wilson_score_interval(p_hat, n, z):
    denominator = 1 + z**2 / n
    centre_adjusted_probability = p_hat + z**2 / (2 * n)
    adjusted_standard_deviation = z * math.sqrt(
        (p_hat * (1 - p_hat) + z**2 / (4 * n)) / n
    )

    lower_bound = (
        centre_adjusted_probability - adjusted_standard_deviation
    ) / denominator
    upper_bound = (
        centre_adjusted_probability + adjusted_standard_deviation
    ) / denominator

    return (lower_bound, upper_bound)


@dataclass
class DataPoint:
    num_qubits_server: int
    makespan_linear: float
    makespan_intlv_1: float
    makespan_intlv_2: float
    improv_1: float
    improv_2: float


@dataclass
class DataMeta:
    timestamp: str
    num_clients: int
    num_iterations: int
    cc: float
    intlv_1_bin_length: float
    intlv_2_bin_length: float
    sim_duration: float


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_clients", "-c", type=int, required=True)
    parser.add_argument("--num_iterations", "-n", type=int, required=True)

    args = parser.parse_args()
    num_clients = args.num_clients
    num_iterations = args.num_iterations

    cc = 1e5
    intlv_1_bin_length = 5e4
    intlv_2_bin_length = 1e5

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    start = time.time()

    data_points: List[DataPoint] = []
    for num_qubits in [2, 5, 10]:
        makespan_linear = bqc_computation(
            num_clients, num_iterations, True, cc, num_qubits, False, 0
        )
        makespan_interleaved_bin_5e4 = bqc_computation(
            num_clients, num_iterations, False, cc, num_qubits, True, intlv_1_bin_length
        )
        makespan_interleaved_bin_1e5 = bqc_computation(
            num_clients, num_iterations, False, cc, num_qubits, True, intlv_2_bin_length
        )

        print(f"# qubits = {num_qubits}:")
        print(f"linear: {makespan_linear:_}")
        print(f"interleaved, bin 5e4: {makespan_interleaved_bin_5e4:_}")
        print(f"interleaved, bin 1e5: {makespan_interleaved_bin_1e5:_}")

        improv_5e4 = round(1 - makespan_interleaved_bin_5e4 / makespan_linear, 3)
        improv_1e5 = round(1 - makespan_interleaved_bin_1e5 / makespan_linear, 3)
        print(f"improvement 5e4: {improv_5e4}")
        print(f"improvement 1e5: {improv_1e5}")

        point = DataPoint(
            num_qubits_server=num_qubits,
            makespan_linear=makespan_linear,
            makespan_intlv_1=makespan_interleaved_bin_5e4,
            makespan_intlv_2=makespan_interleaved_bin_1e5,
            improv_1=improv_5e4,
            improv_2=improv_1e5,
        )
        data_points.append(point)

    end = time.time()
    duration = round(end - start, 2)

    abs_dir = relative_to_cwd(f"data")
    Path(abs_dir).mkdir(parents=True, exist_ok=True)
    last_path = os.path.join(abs_dir, "LAST.json")
    timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")

    meta = DataMeta(
        timestamp=timestamp,
        num_clients=num_clients,
        num_iterations=num_iterations,
        cc=cc,
        intlv_1_bin_length=intlv_1_bin_length,
        intlv_2_bin_length=intlv_2_bin_length,
        sim_duration=duration,
    )
    data = Data(meta=meta, data_points=data_points)
    json_data = asdict(data)

    with open(last_path, "w") as datafile:
        json.dump(json_data, datafile)
    with open(timestamp_path, "w") as datafile:
        json.dump(json_data, datafile)



import itertools
from typing import Dict, FrozenSet, List

from netsquid.components import ClassicalChannel
from netsquid.components.models import FibreDelayModel
from netsquid.components.models.qerrormodels import QuantumErrorModel
from netsquid.components.qprocessor import PhysicalInstruction, QuantumProcessor
from netsquid.nodes.connections import Connection

from qoala.lang.ehi import EhiLinkInfo, EhiNetworkInfo

# Ignore type since whole 'config' module is ignored by mypy
from qoala.runtime.config import ProcNodeConfig, ProcNodeNetworkConfig  # type: ignore
from qoala.runtime.lhi import (
    INSTR_MEASURE_INSTANT,
    LhiLatencies,
    LhiLinkInfo,
    LhiNetworkInfo,
    LhiNetworkSchedule,
    LhiProcNodeInfo,
    LhiTopology,
    LhiTopologyBuilder,
)
from qoala.runtime.lhi_to_ehi import LhiConverter
from qoala.runtime.ntf import NtfInterface
from qoala.sim.entdist.entdist import EntDist
from qoala.sim.entdist.entdistcomp import EntDistComponent
from qoala.sim.network import ProcNodeNetwork
from qoala.sim.procnode import ProcNode


class ClassicalConnection(Connection):
    def __init__(self, name: str, length: float):
        super().__init__(name=name)
        self.add_subcomponent(
            ClassicalChannel(
                "Channel_A2B", length=length, models={"delay_model": FibreDelayModel()}
            )
        )
        self.ports["A"].forward_input(self.subcomponents["Channel_A2B"].ports["send"])
        self.subcomponents["Channel_A2B"].ports["recv"].forward_output(self.ports["B"])


def build_qprocessor_from_topology(
    name: str, topology: LhiTopology
) -> QuantumProcessor:
    num_qubits = len(topology.qubit_infos)

    mem_noise_models: List[QuantumErrorModel] = []
    for i in range(num_qubits):
        info = topology.qubit_infos[i]
        noise_model = info.error_model(**info.error_model_kwargs)
        mem_noise_models.append(noise_model)

    phys_instructions: List[PhysicalInstruction] = []
    # single-qubit gates
    for qubit_id, gate_infos in topology.single_gate_infos.items():
        for gate_info in gate_infos:
            # TODO: refactor this hack
            if gate_info.instruction == INSTR_MEASURE_INSTANT:
                duration = 0.0
            else:
                duration = gate_info.duration

            phys_instr = PhysicalInstruction(
                instruction=gate_info.instruction,
                duration=duration,
                topology=[qubit_id],
                quantum_noise_model=gate_info.error_model(
                    **gate_info.error_model_kwargs
                ),
                parallel=True,
            )
            phys_instructions.append(phys_instr)

    # multi-qubit gates
    for multi_qubit, gate_infos in topology.multi_gate_infos.items():
        qubit_ids = tuple(multi_qubit.qubit_ids)
        for gate_info in gate_infos:
            phys_instr = PhysicalInstruction(
                instruction=gate_info.instruction,
                duration=gate_info.duration,
                topology=[qubit_ids],
                quantum_noise_model=gate_info.error_model(
                    **gate_info.error_model_kwargs
                ),
            )
            phys_instructions.append(phys_instr)

    if topology.all_qubit_gate_infos is not None:
        for gate_info in topology.all_qubit_gate_infos:
            phys_instr = PhysicalInstruction(
                instruction=gate_info.instruction,
                duration=gate_info.duration,
                topology=[tuple(range(num_qubits))],
                quantum_noise_model=gate_info.error_model(
                    **gate_info.error_model_kwargs
                ),
            )
            phys_instructions.append(phys_instr)

    return QuantumProcessor(
        name=name,
        num_positions=num_qubits,
        mem_noise_models=mem_noise_models,
        phys_instructions=phys_instructions,
    )


def build_procnode_from_config(
    cfg: ProcNodeConfig, network_ehi: EhiNetworkInfo
) -> ProcNode:
    topology = LhiTopologyBuilder.from_config(cfg.topology)

    ntf_interface_cls = cfg.ntf.to_ntf_interface()
    ntf_interface = ntf_interface_cls()

    qprocessor = build_qprocessor_from_topology(name=cfg.node_name, topology=topology)
    latencies = LhiLatencies.from_config(cfg.latencies)
    procnode = ProcNode(
        cfg.node_name,
        qprocessor=qprocessor,
        qdevice_topology=topology,
        latencies=latencies,
        ntf_interface=ntf_interface,
        node_id=cfg.node_id,
        network_ehi=network_ehi,
        deterministic_scheduler=cfg.determ_sched,
        use_deadlines=cfg.use_deadlines,
        fcfs=cfg.fcfs,
        prio_epr=cfg.prio_epr,
        is_predictable=cfg.is_predictable,
    )

    # TODO: refactor this hack
    procnode.qnos.processor._latencies.qnos_instr_time = cfg.latencies.qnos_instr_time
    procnode.host.processor._latencies.host_instr_time = cfg.latencies.host_instr_time
    procnode.host.processor._latencies.host_peer_latency = (
        cfg.latencies.host_peer_latency
    )
    return procnode


def build_network_from_config(config: ProcNodeNetworkConfig) -> ProcNodeNetwork:
    procnodes: Dict[str, ProcNode] = {}

    ehi_links: Dict[FrozenSet[int], EhiLinkInfo] = {}
    for link_between_nodes in config.links:
        lhi_link = LhiLinkInfo.from_config(link_between_nodes.link_config)
        ehi_link = LhiConverter.link_info_to_ehi(lhi_link)
        ids = (link_between_nodes.node_id1, link_between_nodes.node_id2)
        node_link = frozenset(ids)
        ehi_links[node_link] = ehi_link
    nodes = {cfg.node_id: cfg.node_name for cfg in config.nodes}
    if config.netschedule is not None:
        lhi_netschedule = LhiNetworkSchedule.from_config(config.netschedule)
        ehi_netschedule = LhiConverter.netschedule_to_ehi(lhi_netschedule)
        network_ehi = EhiNetworkInfo(nodes, ehi_links, ehi_netschedule)
    else:
        network_ehi = EhiNetworkInfo(nodes, ehi_links)

    for cfg in config.nodes:
        procnodes[cfg.node_name] = build_procnode_from_config(cfg, network_ehi)

    ns_nodes = [procnode.node for procnode in procnodes.values()]
    entdistcomp = EntDistComponent(network_ehi)
    entdist = EntDist(nodes=ns_nodes, ehi_network=network_ehi, comp=entdistcomp)

    for link_between_nodes in config.links:
        link = LhiLinkInfo.from_config(link_between_nodes.link_config)
        n1 = link_between_nodes.node_id1
        n2 = link_between_nodes.node_id2
        entdist.add_sampler(n1, n2, link)

    def get_latency(node1: int, node2: int) -> float:
        if config.cconns is None:
            return 0.0
        for cconn in config.cconns:
            if {cconn.node_id1, cconn.node_id2} == {node1, node2}:
                return cconn.latency  # type: ignore
        return 0.0

    for s1, s2 in itertools.combinations(procnodes.values(), 2):
        latency = get_latency(s1.node.node_id, s2.node.node_id)
        s1.connect_to(s2, latency)

    # TODO: make this configurable
    node_entdist_latency = 0
    for name, procnode in procnodes.items():
        chan_ne = ClassicalChannel(f"chan_{name}_entdist", delay=node_entdist_latency)
        procnode.node.entdist_out_port.connect(chan_ne.ports["send"])
        chan_ne.ports["recv"].connect(entdistcomp.node_in_port(name))

        chan_en = ClassicalChannel(f"chan_entdist_{name}", delay=node_entdist_latency)
        entdistcomp.node_out_port(name).connect(chan_en.ports["send"])
        chan_en.ports["recv"].connect(procnode.node.entdist_in_port)

    return ProcNodeNetwork(procnodes, entdist)


def build_procnode_from_lhi(
    id: int,
    name: str,
    topology: LhiTopology,
    latencies: LhiLatencies,
    network_lhi: LhiNetworkInfo,
    ntf: NtfInterface,
) -> ProcNode:
    qprocessor = build_qprocessor_from_topology(f"{name}_processor", topology)
    network_ehi = LhiConverter.network_to_ehi(network_lhi)
    return ProcNode(
        name=name,
        node_id=id,
        qprocessor=qprocessor,
        qdevice_topology=topology,
        latencies=latencies,
        ntf_interface=ntf,
        network_ehi=network_ehi,
    )


def build_network_from_lhi(
    procnode_infos: List[LhiProcNodeInfo],
    ntfs: List[NtfInterface],
    network_lhi: LhiNetworkInfo,
) -> ProcNodeNetwork:
    procnodes: Dict[str, ProcNode] = {}

    # TODO: refactor two separate lists (infos and ntfs)
    for info, ntf in zip(procnode_infos, ntfs):
        procnode = build_procnode_from_lhi(
            info.id, info.name, info.topology, info.latencies, network_lhi, ntf
        )
        procnodes[info.name] = procnode

    ns_nodes = [procnode.node for procnode in procnodes.values()]
    network_ehi = LhiConverter.network_to_ehi(network_lhi)
    entdistcomp = EntDistComponent(network_ehi)
    entdist = EntDist(nodes=ns_nodes, ehi_network=network_ehi, comp=entdistcomp)

    for ([n1, n2], link_info) in network_lhi.links.items():
        entdist.add_sampler(n1, n2, link_info)

    for s1, s2 in itertools.combinations(procnodes.values(), 2):
        s1.connect_to(s2)

    for name, procnode in procnodes.items():
        procnode.node.entdist_out_port.connect(entdistcomp.node_in_port(name))
        procnode.node.entdist_in_port.connect(entdistcomp.node_out_port(name))

    return ProcNodeNetwork(procnodes, entdist)



from __future__ import annotations

from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

from netqasm.lang import operand
from netqasm.lang.encoding import RegisterName
from netqasm.sdk.shared_memory import RegisterGroup, setup_registers

from qoala.lang.request import RequestRoutine
from qoala.lang.routine import LocalRoutine
from qoala.runtime.sharedmem import MemAddr, SharedMemory


class RegisterMeta:
    @classmethod
    def prefixes(cls) -> List[str]:
        return ["R", "C", "Q", "M"]

    @classmethod
    def parse(cls, name: str) -> Tuple[RegisterName, int]:
        assert len(name) >= 2
        assert name[0] in cls.prefixes()
        group = RegisterName[name[0]]
        index = int(name[1:])
        assert index < 16
        return group, index


class HostMemory:
    """Classical program memory only available to the Host.
    Simple mapping from variable names to values."""

    def __init__(self, pid: int) -> None:
        self._pid = pid

        # Host memory is represented as a mapping from variables to values.
        # Variables have a name (str) and values (int).
        self._mem: Dict[str, int] = {}

        # Vectors are stored separately.
        self._vec_mem: Dict[str, List[int]] = {}

    def write(self, loc: str, value: int) -> None:
        self._mem[loc] = value

    def read(self, loc: str) -> int:
        return self._mem[loc]

    def write_vec(self, loc: str, values: List[int]) -> None:
        self._vec_mem[loc] = values

    def read_vec(self, loc: str) -> List[int]:
        return self._vec_mem[loc]


@dataclass
class RunningLocalRoutine:
    routine: LocalRoutine
    params_addr: MemAddr
    result_addr: MemAddr


@dataclass
class RunningRequestRoutine:
    routine: RequestRoutine
    params_addr: MemAddr
    result_addr: MemAddr
    cb_input_addrs: List[MemAddr]
    cb_output_addrs: List[MemAddr]


class QnosMemory:
    """Classical program memory only available to Qnos."""

    def __init__(self, pid: int) -> None:
        self._pid = pid

        # TODO: allow multiple instances of same routine (name)?
        # Currently not possible
        self._running_local_routines: Dict[str, RunningLocalRoutine] = {}
        self._running_request_routines: Dict[str, RunningRequestRoutine] = {}

        # NetQASM registers.
        register_names: Dict[RegisterName, RegisterGroup] = setup_registers()
        self._registers: Dict[RegisterName, Dict[int, int]] = {}
        # TODO fix this abomination of handling registers
        for name in register_names.keys():
            self._registers[name] = {}  # type: ignore
            for i in range(16):
                self._registers[name][i] = 0  # type: ignore

    def add_running_local_routine(self, routine: RunningLocalRoutine) -> None:
        self._running_local_routines[routine.routine.name] = routine

    def get_running_local_routine(self, name: str) -> RunningLocalRoutine:
        return self._running_local_routines[name]

    def get_all_running_local_routines(self) -> Dict[str, RunningLocalRoutine]:
        return self._running_local_routines

    def add_running_request_routine(self, routine: RunningRequestRoutine) -> None:
        self._running_request_routines[routine.routine.name] = routine

    def get_running_request_routine(self, name: str) -> RunningRequestRoutine:
        return self._running_request_routines[name]

    def get_all_running_request_routines(self) -> Dict[str, RunningRequestRoutine]:
        return self._running_request_routines

    def set_reg_value(self, register: Union[str, operand.Register], value: int) -> None:
        if isinstance(register, str):
            name, index = RegisterMeta.parse(register)
        else:
            name, index = register.name, register.index
        self._registers[name][index] = value  # type: ignore

    def get_reg_value(self, register: Union[str, operand.Register]) -> int:
        if isinstance(register, str):
            name, index = RegisterMeta.parse(register)
        else:
            name, index = register.name, register.index
        return self._registers[name][index]  # type: ignore

    # for compatibility with netqasm Futures
    def get_register(self, register: Union[str, operand.Register]) -> Optional[int]:
        return self.get_reg_value(register)


class ProgramMemory:
    """Dynamic runtime memory, divided into
    - Host Memory: local to the Host
    - Qnos Memory: local to Qnos
    - Shared Memory: shared between Host, Qnos and Netstack"""

    def __init__(self, pid: int) -> None:
        self._pid: int = pid

        # TODO: remove pids?
        self._host_memory = HostMemory(pid)
        self._qnos_memory = QnosMemory(pid)
        self._shared_memory = SharedMemory()

    @property
    def host_mem(self) -> HostMemory:
        return self._host_memory

    @property
    def shared_mem(self) -> SharedMemory:
        return self._shared_memory

    @property
    def qnos_mem(self) -> QnosMemory:
        return self._qnos_memory



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Optional

import netsquid as ns

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NetworkScheduleConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.logging import LogManager
from qoala.util.runner import run_two_node_app


def create_procnode_cfg(name: str, id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
        latencies=LatenciesConfig(qnos_instr_time=1000),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


@dataclass
class QkdResult:
    alice_result: BatchResult
    bob_result: BatchResult


def run_qkd(
    num_iterations: int,
    alice_file: str,
    bob_file: str,
    num_pairs: Optional[int] = None,
):
    num_qubits = 3
    alice_id = 0
    bob_id = 1

    alice_node_cfg = create_procnode_cfg("alice", alice_id, num_qubits)
    bob_node_cfg = create_procnode_cfg("bob", bob_id, num_qubits)

    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )
    pattern = [(alice_id, i, bob_id, i) for i in range(num_iterations)]
    network_cfg.netschedule = NetworkScheduleConfig(
        bin_length=1_500, first_bin=0, bin_pattern=pattern, repeat_period=20_000
    )

    alice_program = load_program(alice_file)
    bob_program = load_program(bob_file)

    if num_pairs is not None:
        alice_input = ProgramInput({"bob_id": bob_id, "N": num_pairs})
        bob_input = ProgramInput({"alice_id": alice_id, "N": num_pairs})
    else:
        alice_input = ProgramInput({"bob_id": bob_id})
        bob_input = ProgramInput({"alice_id": alice_id})

    app_result = run_two_node_app(
        num_iterations=num_iterations,
        programs={"alice": alice_program, "bob": bob_program},
        program_inputs={"alice": alice_input, "bob": bob_input},
        network_cfg=network_cfg,
        linear=True,
    )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]

    return QkdResult(alice_result, bob_result)


def qkd_1pair_md():
    ns.sim_reset()
    # LogManager.enable_task_logger(True)
    LogManager.set_log_level("WARNING")

    num_iterations = 10
    alice_file = "qkd_1pair_MD_alice.iqoala"
    bob_file = "qkd_1pair_MD_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    print(alice_outcomes)

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        assert alice["m0"] == bob["m0"]


def qkd_1pair_ck():
    ns.sim_reset()

    num_iterations = 10
    alice_file = "qkd_1pair_CK_alice.iqoala"
    bob_file = "qkd_1pair_CK_bob.iqoala"

    qkd_result = run_qkd(num_iterations, alice_file, bob_file)
    alice_results = qkd_result.alice_result.results
    bob_results = qkd_result.bob_result.results

    assert len(alice_results) == num_iterations
    assert len(bob_results) == num_iterations

    alice_outcomes = [alice_results[i].values for i in range(num_iterations)]
    bob_outcomes = [bob_results[i].values for i in range(num_iterations)]

    for alice, bob in zip(alice_outcomes, bob_outcomes):
        assert alice["m0"] == bob["m0"]


def test_qkd_1pair_md():
    qkd_1pair_md()


def test_qkd_1pair_ck():
    qkd_1pair_ck()


if __name__ == "__main__":
    test_qkd_1pair_md()
    # test_qkd_1pair_ck()



from __future__ import annotations

from dataclasses import dataclass
from typing import Generator

from pydynaa import EventExpression
from qoala.lang.ehi import EhiNetworkInfo
from qoala.runtime.message import Message
from qoala.sim.componentprot import ComponentProtocol, PortListener
from qoala.sim.events import (
    EVENT_WAIT,
    SIGNAL_ENTD_NSTK_MSG,
    SIGNAL_HOST_NSTK_MSG,
    SIGNAL_MEMORY_FREED,
    SIGNAL_NSTK_NSTK_MSG,
    SIGNAL_QNOS_NSTK_MSG,
)
from qoala.sim.memmgr import MemoryManager
from qoala.sim.netstack.netstackcomp import NetstackComponent
from qoala.sim.qdevice import QDevice


@dataclass
class NetstackLatencies:
    netstack_peer_latency: float = 0  # processing time for messages from remote node

    @classmethod
    def all_zero(cls) -> NetstackLatencies:
        # NOTE: can also just use NetstackLatencies() which will default all values to 0
        # However, using this classmethod makes this behavior more explicit and clear.
        return NetstackLatencies(0)


class NetstackInterface(ComponentProtocol):
    """NetSquid protocol representing the QNodeOS network stack."""

    def __init__(
        self,
        comp: NetstackComponent,
        ehi_network: EhiNetworkInfo,
        qdevice: QDevice,
        memmgr: MemoryManager,
    ) -> None:
        """Network stack protocol constructor. Typically created indirectly through
        constructing a `Qnos` instance.

        :param comp: NetSquid component representing the network stack
        :param qnos: `Qnos` protocol that owns this protocol
        """
        super().__init__(name=f"{comp.name}_protocol", comp=comp)
        self._comp = comp
        self._qdevice = qdevice
        self._ehi_network = ehi_network
        self._memmgr = memmgr

        self.add_listener(
            "host",
            PortListener(self._comp.host_in_port, SIGNAL_HOST_NSTK_MSG),
        )

        self.add_listener(
            "qnos",
            PortListener(self._comp.qnos_in_port, SIGNAL_QNOS_NSTK_MSG),
        )

        self.add_listener(
            "qnos_mem",
            PortListener(self._comp.qnos_mem_in_port, SIGNAL_MEMORY_FREED),
        )

        self.add_listener(
            "entdist",
            PortListener(self._comp.entdist_in_port, SIGNAL_ENTD_NSTK_MSG),
        )

        for peer in self._ehi_network.nodes.values():
            if peer == self._comp.node_name:
                continue
            self.add_listener(
                f"peer_{peer}",
                PortListener(
                    self._comp.peer_in_port(peer), f"{SIGNAL_NSTK_NSTK_MSG}_{peer}"
                ),
            )

    def send_entdist_msg(self, msg: Message) -> None:
        """Send a message to the Entdist."""
        self._comp.entdist_out_port.tx_output(msg)

    def receive_entdist_msg(self) -> Generator[EventExpression, None, Message]:
        """Receive a message from the Entdist. Block until there is at least one
        message."""
        yield from self._wait_for_msg("entdist", SIGNAL_ENTD_NSTK_MSG)
        return self._pop_any_msg("entdist")

    def send_peer_msg(self, peer: str, msg: Message) -> None:
        """Send a message to the network stack of the other node."""

        self._comp.peer_out_port(peer).tx_output(msg)

    def receive_peer_msg(self, peer: str) -> Generator[EventExpression, None, Message]:
        """Receive a message from the network stack of the other node. Block until
        there is at least one message."""

        yield from self._wait_for_msg(f"peer_{peer}", f"{SIGNAL_NSTK_NSTK_MSG}_{peer}")
        return self._pop_any_msg(f"peer_{peer}")

    @property
    def qdevice(self) -> QDevice:
        return self._qdevice

    @property
    def node_id(self) -> int:
        return self._qdevice.node.ID  # type: ignore

    @property
    def memmgr(self) -> MemoryManager:
        return self._memmgr

    def remote_id_to_peer_name(self, remote_id: int) -> str:
        return self._ehi_network.nodes[remote_id]

    def wait(self, delta_time: float) -> Generator[EventExpression, None, None]:
        self._schedule_after(delta_time, EVENT_WAIT)
        event_expr = EventExpression(source=self, event_type=EVENT_WAIT)
        yield event_expr



from abc import ABC, abstractmethod
from typing import Dict, List, Type

from netqasm.lang.instr import core, nv, trapped_ion, vanilla
from netqasm.lang.instr.base import NetQASMInstruction
from netqasm.lang.instr.flavour import (
    Flavour,
    NVFlavour,
    TrappedIonFlavour,
    VanillaFlavour,
)
from netsquid.components.instructions import (
    INSTR_CNOT,
    INSTR_CXDIR,
    INSTR_CYDIR,
    INSTR_CZ,
    INSTR_H,
    INSTR_INIT,
    INSTR_MEASURE,
    INSTR_ROT_X,
    INSTR_ROT_Y,
    INSTR_ROT_Z,
    INSTR_X,
    INSTR_Y,
    INSTR_Z,
)
from netsquid.components.instructions import Instruction as NetSquidInstruction

from qoala.runtime.instructions import (
    INSTR_BICHROMATIC,
    INSTR_MEASURE_ALL,
    INSTR_ROT_X_ALL,
    INSTR_ROT_Y_ALL,
    INSTR_ROT_Z_ALL,
)
from qoala.runtime.lhi import INSTR_MEASURE_INSTANT


class NtfInterface(ABC):
    @abstractmethod
    def flavour(self) -> Type[Flavour]:
        raise NotImplementedError

    @abstractmethod
    def native_to_netqasm(
        self, ns_instr: Type[NetSquidInstruction]
    ) -> List[Type[NetQASMInstruction]]:
        """Responsiblity of implementor that return instructions are of the
        flavour returned by flavour()."""
        raise NotImplementedError

    @abstractmethod
    def netqasm_to_native(
        self, nq_instr: Type[NetQASMInstruction]
    ) -> List[Type[NetSquidInstruction]]:
        raise NotImplementedError


class GenericNtf(NtfInterface):
    _NS_NQ_MAP: Dict[Type[NetSquidInstruction], List[Type[NetQASMInstruction]]] = {
        INSTR_INIT: [core.InitInstruction],
        INSTR_X: [vanilla.GateXInstruction],
        INSTR_Y: [vanilla.GateYInstruction],
        INSTR_Z: [vanilla.GateZInstruction],
        INSTR_H: [vanilla.GateHInstruction],
        INSTR_ROT_X: [vanilla.RotXInstruction],
        INSTR_ROT_Y: [vanilla.RotYInstruction],
        INSTR_ROT_Z: [vanilla.RotZInstruction],
        INSTR_CNOT: [vanilla.CnotInstruction],
        INSTR_CZ: [vanilla.CphaseInstruction],
        INSTR_MEASURE: [core.MeasInstruction],
        INSTR_MEASURE_INSTANT: [core.MeasInstruction],
    }

    _NQ_NS_MAP: Dict[Type[NetSquidInstruction], List[Type[NetQASMInstruction]]] = {
        core.InitInstruction: [INSTR_INIT],
        vanilla.GateXInstruction: [INSTR_X],
        vanilla.GateYInstruction: [INSTR_Y],
        vanilla.GateZInstruction: [INSTR_Z],
        vanilla.GateHInstruction: [INSTR_H],
        vanilla.RotXInstruction: [INSTR_ROT_X],
        vanilla.RotYInstruction: [INSTR_ROT_Y],
        vanilla.RotZInstruction: [INSTR_ROT_Z],
        vanilla.CnotInstruction: [INSTR_CNOT],
        vanilla.CphaseInstruction: [INSTR_CZ],
        core.MeasInstruction: [INSTR_MEASURE],
        core.MeasInstruction: [INSTR_MEASURE_INSTANT],
    }

    def flavour(self) -> Type[Flavour]:
        return VanillaFlavour  # type: ignore

    def native_to_netqasm(
        self, ns_instr: Type[NetSquidInstruction]
    ) -> List[Type[NetQASMInstruction]]:
        return self._NS_NQ_MAP[ns_instr]

    def netqasm_to_native(
        self, nq_instr: Type[NetQASMInstruction]
    ) -> List[Type[NetSquidInstruction]]:
        return self._NQ_NS_MAP[nq_instr]


class NvNtf(NtfInterface):
    _NS_NQ_MAP: Dict[Type[NetSquidInstruction], List[Type[NetQASMInstruction]]] = {
        INSTR_INIT: [core.InitInstruction],
        INSTR_ROT_X: [nv.RotXInstruction],
        INSTR_ROT_Y: [nv.RotYInstruction],
        INSTR_ROT_Z: [nv.RotZInstruction],
        INSTR_CXDIR: [nv.ControlledRotXInstruction],
        INSTR_CYDIR: [nv.ControlledRotYInstruction],
        INSTR_MEASURE: [core.MeasInstruction],
        INSTR_MEASURE_INSTANT: [core.MeasInstruction],
    }

    _NQ_NS_MAP: Dict[Type[NetQASMInstruction], List[Type[NetSquidInstruction]]] = {
        core.InitInstruction: [INSTR_INIT],
        nv.RotXInstruction: [INSTR_ROT_X],
        nv.RotYInstruction: [INSTR_ROT_Y],
        nv.RotZInstruction: [INSTR_ROT_Z],
        nv.ControlledRotXInstruction: [INSTR_CXDIR],
        nv.ControlledRotYInstruction: [INSTR_CYDIR],
        core.MeasInstruction: [INSTR_MEASURE],
        core.MeasInstruction: [INSTR_MEASURE_INSTANT],
    }

    def flavour(self) -> Type[Flavour]:
        return NVFlavour  # type: ignore

    def native_to_netqasm(
        self, ns_instr: Type[NetSquidInstruction]
    ) -> List[Type[NetQASMInstruction]]:
        return self._NS_NQ_MAP[ns_instr]

    def netqasm_to_native(
        self, nq_instr: Type[NetQASMInstruction]
    ) -> List[Type[NetSquidInstruction]]:
        return self._NQ_NS_MAP[nq_instr]


class TrappedIonNtf(NtfInterface):
    #  TODO Check if this is correct
    _NS_NQ_MAP: Dict[Type[NetSquidInstruction], List[Type[NetQASMInstruction]]] = {
        INSTR_INIT: [core.InitInstruction, trapped_ion.AllQubitsInitInstruction],
        INSTR_ROT_Z: [trapped_ion.RotZInstruction],
        INSTR_BICHROMATIC: [trapped_ion.BichromaticInstruction],
        INSTR_MEASURE_ALL: [trapped_ion.AllQubitsMeasInstruction],
        INSTR_ROT_X_ALL: [trapped_ion.AllQubitsRotXInstruction],
        INSTR_ROT_Y_ALL: [trapped_ion.AllQubitsRotYInstruction],
        INSTR_ROT_Z_ALL: [trapped_ion.AllQubitsRotZInstruction],
        INSTR_MEASURE: [core.MeasInstruction],
        INSTR_MEASURE_INSTANT: [core.MeasInstruction],
    }

    _NQ_NS_MAP: Dict[Type[NetSquidInstruction], List[Type[NetQASMInstruction]]] = {
        core.InitInstruction: [INSTR_INIT],
        trapped_ion.AllQubitsInitInstruction: [INSTR_INIT],
        trapped_ion.RotZInstruction: [INSTR_ROT_Z],
        core.MeasInstruction: [INSTR_MEASURE],
        trapped_ion.AllQubitsMeasInstruction: [INSTR_MEASURE_ALL],
        trapped_ion.AllQubitsRotXInstruction: [INSTR_ROT_X_ALL],
        trapped_ion.AllQubitsRotYInstruction: [INSTR_ROT_Y_ALL],
        trapped_ion.AllQubitsRotZInstruction: [INSTR_ROT_Z_ALL],
        trapped_ion.BichromaticInstruction: [INSTR_BICHROMATIC],
    }

    def flavour(self) -> Type[Flavour]:
        return TrappedIonFlavour  # type: ignore

    def native_to_netqasm(
        self, ns_instr: Type[NetSquidInstruction]
    ) -> List[Type[NetQASMInstruction]]:
        return self._NS_NQ_MAP[ns_instr]

    def netqasm_to_native(
        self, nq_instr: Type[NetQASMInstruction]
    ) -> List[Type[NetSquidInstruction]]:
        return self._NQ_NS_MAP[nq_instr]


class NtfInterfaceConfigInterface(ABC):
    @abstractmethod
    def to_ntf_interface(self) -> Type[NtfInterface]:
        raise NotImplementedError



from __future__ import annotations

import os
from dataclasses import dataclass
from typing import Dict, List

import netsquid as ns

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchInfo, BatchResult, ProgramBatch, ProgramInput
from qoala.sim.build import build_network_from_config
from qoala.sim.network import ProcNodeNetwork


def topology_config(num_qubits: int) -> TopologyConfig:
    return TopologyConfig.perfect_config_uniform(
        num_qubits,
        single_instructions=[
            "INSTR_INIT",
            "INSTR_ROT_X",
            "INSTR_ROT_Y",
            "INSTR_ROT_Z",
            "INSTR_X",
            "INSTR_Y",
            "INSTR_Z",
            "INSTR_H",
            "INSTR_MEASURE",
        ],
        single_duration=1e3,
        two_instructions=["INSTR_CNOT", "INSTR_CZ"],
        two_duration=100e3,
    )


def get_client_config(id: int) -> ProcNodeConfig:
    # client only needs 1 qubit
    return ProcNodeConfig(
        node_name=f"client_{id}",
        node_id=id,
        topology=topology_config(1),
        latencies=LatenciesConfig(
            host_instr_time=500, host_peer_latency=30_000, qnos_instr_time=1000
        ),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def get_server_config(id: int, num_qubits: int) -> ProcNodeConfig:
    return ProcNodeConfig(
        node_name="server",
        node_id=id,
        topology=topology_config(num_qubits),
        latencies=LatenciesConfig(
            host_instr_time=500, host_peer_latency=30_000, qnos_instr_time=1000
        ),
        ntf=NtfConfig.from_cls_name("GenericNtf"),
    )


def create_network(
    server_cfg: ProcNodeConfig,
    client_configs: List[ProcNodeConfig],
    num_clients: int,
) -> ProcNodeNetwork:
    assert len(client_configs) == num_clients

    node_cfgs = [server_cfg] + client_configs
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=node_cfgs, link_duration=1000
    )
    return build_network_from_config(network_cfg)


@dataclass
class BqcResult:
    client_batches: List[Dict[int, ProgramBatch]]
    client_results: List[Dict[int, BatchResult]]


def load_server_program(remote_name: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), "vbqc_server.iqoala")
    with open(path) as file:
        server_text = file.read()
    program = QoalaParser(server_text).parse()

    # Replace "client" by e.g. "client_1"
    program.meta.csockets[0] = remote_name
    program.meta.epr_sockets[0] = remote_name

    return program


def load_client_program() -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), "vbqc_client.iqoala")
    with open(path) as file:
        client_text = file.read()
    return QoalaParser(client_text).parse()


def create_server_batch(
    client_id: int,
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
    deadline: int,
) -> BatchInfo:
    server_program = load_server_program(remote_name=f"client_{client_id}")
    return BatchInfo(
        program=server_program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=deadline,
    )


def create_client_batch(
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
    deadline: int,
) -> BatchInfo:
    client_program = load_client_program()
    return BatchInfo(
        program=client_program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=deadline,
    )


def run_bqc(
    alpha,
    beta,
    theta1,
    theta2,
    dummy0,
    dummy1,
    num_iterations: List[int],
    deadlines: List[int],
    num_clients: int,
):
    ns.sim_reset()

    # server needs to have 2 qubits per client
    server_num_qubits = num_clients * 2
    server_config = get_server_config(id=0, num_qubits=server_num_qubits)
    client_configs = [get_client_config(i) for i in range(1, num_clients + 1)]

    network = create_network(server_config, client_configs, num_clients)
    server_procnode = network.nodes["server"]
    server_batches: Dict[int, ProgramBatch] = {}  # client ID -> server batch
    client_batches: Dict[int, ProgramBatch] = {}  # client ID -> client batch

    for client_id in range(1, num_clients + 1):
        # index in num_iterations and deadlines list
        index = client_id - 1

        server_inputs = [
            ProgramInput({"client_id": client_id}) for _ in range(num_iterations[index])
        ]

        server_unit_module = UnitModule.from_full_ehi(server_procnode.memmgr.get_ehi())
        server_batch_info = create_server_batch(
            client_id=client_id,
            inputs=server_inputs,
            unit_module=server_unit_module,
            num_iterations=num_iterations[index],
            deadline=deadlines[index],
        )

        server_batches[client_id] = server_procnode.submit_batch(server_batch_info)

    for client_id in range(1, num_clients + 1):
        # index in num_iterations and deadlines list
        index = client_id - 1

        client_inputs = [
            ProgramInput(
                {
                    "server_id": 0,
                    "alpha": alpha,
                    "beta": beta,
                    "theta1": theta1,
                    "theta2": theta2,
                    "dummy0": dummy0,
                    "dummy1": dummy1,
                }
            )
            for _ in range(num_iterations[index])
        ]

        client_procnode = network.nodes[f"client_{client_id}"]

        client_unit_module = UnitModule.from_full_ehi(client_procnode.memmgr.get_ehi())
        client_batch_info = create_client_batch(
            client_inputs, client_unit_module, num_iterations[index], deadlines[index]
        )

        client_batches[client_id] = client_procnode.submit_batch(client_batch_info)
        batch_id = client_batches[client_id].batch_id
        server_pids = [inst.pid for inst in server_batches[client_id].instances]
        print(
            f"client ID: {client_id}, batch ID: {batch_id}, server PIDs: {server_pids}"
        )
        client_procnode.initialize_processes(
            remote_pids={batch_id: server_pids}, linear=True
        )

    client_pids = {
        server_batches[client_id].batch_id: [
            inst.pid for inst in client_batches[client_id].instances
        ]
        for client_id in range(1, num_clients + 1)
    }
    print(f"client PIDs: {client_pids}")
    server_procnode.initialize_processes(remote_pids=client_pids, linear=True)

    network.start()
    start_time = ns.sim_time()
    ns.sim_run()
    end_time = ns.sim_time()
    makespan = end_time - start_time

    client_procnodes = [network.nodes[f"client_{i}"] for i in range(1, num_clients + 1)]
    client_batches = [node.get_batches() for node in client_procnodes]

    client_results: List[Dict[int, BatchResult]]
    client_results = [node.scheduler.get_batch_results() for node in client_procnodes]

    return BqcResult(client_batches, client_results), makespan


def check_computation(
    alpha,
    beta,
    theta1,
    theta2,
    dummy0,
    dummy1,
    expected,
    num_iterations,
    deadlines,
    num_clients,
):
    ns.sim_reset()
    bqc_result, makespan = run_bqc(
        alpha=alpha,
        beta=beta,
        theta1=theta1,
        theta2=theta2,
        dummy0=dummy0,
        dummy1=dummy1,
        num_iterations=num_iterations,
        deadlines=deadlines,
        num_clients=num_clients,
    )

    batch_success_probabilities: List[float] = []

    for i in range(num_clients):
        assert len(bqc_result.client_results[i]) == 1
        batch_result = bqc_result.client_results[i][0]
        assert len(bqc_result.client_batches[i]) == 1
        program_batch = bqc_result.client_batches[i][0]
        batch_iterations = program_batch.info.num_iterations

        m2s = [result.values["m2"] for result in batch_result.results]
        correct_outcomes = len([m2 for m2 in m2s if m2 == expected])
        succ_prob = round(correct_outcomes / batch_iterations, 2)
        batch_success_probabilities.append(succ_prob)

    return batch_success_probabilities, makespan


def compute_succ_prob_computation(
    num_clients: int,
    num_iterations: List[int],
    deadlines: List[int],
):
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    return check_computation(
        alpha=8,
        beta=24,
        theta1=2,
        theta2=22,
        dummy0=0,
        dummy1=0,
        expected=1,
        num_iterations=num_iterations,
        deadlines=deadlines,
        num_clients=num_clients,
    )


def check_trap(
    alpha,
    beta,
    theta1,
    theta2,
    dummy0,
    dummy1,
    num_iterations,
    deadlines,
    num_clients,
):
    ns.sim_reset()
    bqc_result, makespan = run_bqc(
        alpha=alpha,
        beta=beta,
        theta1=theta1,
        theta2=theta2,
        dummy0=dummy0,
        dummy1=dummy1,
        num_iterations=num_iterations,
        deadlines=deadlines,
        num_clients=num_clients,
    )

    batch_success_probabilities: List[float] = []

    for i in range(num_clients):
        assert len(bqc_result.client_results[i]) == 1
        batch_result = bqc_result.client_results[i][0]
        assert len(bqc_result.client_batches[i]) == 1
        program_batch = bqc_result.client_batches[i][0]
        batch_iterations = program_batch.info.num_iterations

        p1s = [result.values["p1"] for result in batch_result.results]
        p2s = [result.values["p2"] for result in batch_result.results]
        m1s = [result.values["m1"] for result in batch_result.results]
        m2s = [result.values["m2"] for result in batch_result.results]

        if dummy0 == 0:
            # corresponds to "dummy = 1"
            # do normal rotations on qubit 0
            # no rotations on qubit 1
            num_fails = len([(p, m) for (p, m) in zip(p1s, m2s) if p != m])
        else:  # dummy0 = 1
            # corresponds to "dummy = 2"
            # no rotations on qubit 0
            # do normal rotations on qubit 1
            num_fails = len([(p, m) for (p, m) in zip(p2s, m1s) if p != m])

        frac_fail = round(num_fails / batch_iterations, 2)
        batch_success_probabilities.append(1 - frac_fail)

    return batch_success_probabilities, makespan


def compute_succ_prob_trap(
    num_clients: int,
    num_iterations: List[int],
    deadlines: List[int],
):
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    return check_trap(
        alpha=8,
        beta=24,
        theta1=2,
        theta2=22,
        dummy0=0,
        dummy1=1,
        num_iterations=num_iterations,
        deadlines=deadlines,
        num_clients=num_clients,
    )


def bqc_computation(num_clients: int, num_iterations: int):
    succ_probs, makespan = compute_succ_prob_computation(
        num_clients=num_clients,
        num_iterations=[num_iterations] * num_clients,
        deadlines=[1e9] * num_clients,
    )
    print(f"success probabilities: {succ_probs}")
    print(f"makespan: {makespan}")


def bqc_trap(num_clients: int, num_iterations: int):
    succ_probs, makespan = compute_succ_prob_trap(
        num_clients=num_clients,
        num_iterations=[num_iterations] * num_clients,
        deadlines=[1e9] * num_clients,
    )
    print(f"success probabilities: {succ_probs}")
    print(f"makespan: {makespan}")


def test_bqc_1_client_computation():
    bqc_computation(1, 10)


def test_bqc_1_client_trap():
    bqc_trap(1, 10)


def test_bqc_multiple_clients_computation():
    bqc_computation(3, 30)


def test_bqc_multiple_clients_trap():
    bqc_trap(3, 30)


if __name__ == "__main__":
    test_bqc_1_client_computation()
    test_bqc_1_client_trap()
    test_bqc_multiple_clients_computation()
    test_bqc_multiple_clients_trap()



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional

import numpy as np
from netsquid.qubits import qubitapi
from netsquid.qubits.qubit import Qubit

if TYPE_CHECKING:
    from qoala.sim.network import ProcNodeNetwork

T_QubitData = Dict[str, Dict[int, Qubit]]
T_StateData = Dict[str, Dict[int, np.ndarray]]


class GlobalSimData:
    _NETWORK: Optional[ProcNodeNetwork] = None
    _BREAKPOINT_STATES: List[T_StateData] = []

    def set_network(self, network: ProcNodeNetwork) -> None:
        self._NETWORK = network

    def get_network(self) -> Optional[ProcNodeNetwork]:
        return self._NETWORK

    def get_quantum_state(self, save: bool = False) -> T_QubitData:
        network = self.get_network()
        assert network is not None

        qubits: T_QubitData = {}
        states: T_StateData = {}
        for name, qdevice in network.qdevices.items():
            qubits[name] = {}
            states[name] = {}
            for i in range(qdevice.num_positions):
                if qdevice.mem_positions[i].in_use:
                    [q] = qdevice.peek(i)
                    qubits[name][i] = q
                    if save:
                        states[name][i] = qubitapi.reduced_dm(q)
        if save:
            self._BREAKPOINT_STATES.append(states)
        return qubits

    def get_last_breakpoint_state(self) -> T_StateData:
        assert len(self._BREAKPOINT_STATES) > 0
        return self._BREAKPOINT_STATES[-1]



from __future__ import annotations

import json
import math
import os
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from datetime import datetime
from pathlib import Path
from typing import List

import netsquid as ns

from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    ClassicalConnectionConfig,
    LatenciesConfig,
    NetworkScheduleConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchResult, ProgramInput
from qoala.util.runner import run_two_node_app_separate_inputs_plus_local_program


def create_procnode_cfg(
    name: str,
    id: int,
    t1: float,
    t2: float,
    determ: bool,
    deadlines: bool,
    num_qubits: int,
) -> ProcNodeConfig:
    topology = TopologyConfig.uniform_t1t2_qubits_perfect_gates_default_params(
        num_qubits, t1, t2
    )
    if name == "alice":
        latencies = LatenciesConfig(qnos_instr_time=1000, host_instr_time=1000)
    elif name == "bob":
        latencies = LatenciesConfig(qnos_instr_time=100_000, host_instr_time=1000)

    return ProcNodeConfig(
        node_name=name,
        node_id=id,
        topology=topology,
        latencies=latencies,
        ntf=NtfConfig.from_cls_name("GenericNtf"),
        determ_sched=determ,
        use_deadlines=deadlines,
    )


def load_program(path: str) -> QoalaProgram:
    path = os.path.join(os.path.dirname(__file__), path)
    with open(path) as file:
        text = file.read()
    return QoalaParser(text).parse()


@dataclass
class TeleportResult:
    alice_result: BatchResult
    bob_result: BatchResult
    local_result: BatchResult
    total_duration: float


def run_apps(
    num_teleport: int,
    num_local: int,
    num_qubits_bob: int,
    t1: float,
    t2: float,
    cc_latency: float,
    network_bin_len: int,
    network_period: int,
    network_first_bin: int,
    local_busy_duration: int,
) -> TeleportResult:
    ns.sim_reset()

    alice_id = 1
    bob_id = 0

    alice_node_cfg = create_procnode_cfg(
        "alice",
        alice_id,
        t1,
        t2,
        determ=True,
        deadlines=True,
        num_qubits=20,
    )
    bob_node_cfg = create_procnode_cfg(
        "bob",
        bob_id,
        t1,
        t2,
        determ=True,
        deadlines=True,
        num_qubits=num_qubits_bob,
    )

    cconn = ClassicalConnectionConfig.from_nodes(alice_id, bob_id, cc_latency)
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=[alice_node_cfg, bob_node_cfg], link_duration=1000
    )
    pattern = [(alice_id, i, bob_id, i) for i in range(num_teleport)]
    network_cfg.netschedule = NetworkScheduleConfig(
        bin_length=network_bin_len,
        first_bin=network_first_bin,
        bin_pattern=pattern,
        repeat_period=network_period,
    )
    network_cfg.cconns = [cconn]

    alice_program = load_program("programs/teleport_alice.iqoala")
    bob_program = load_program("programs/teleport_bob.iqoala")
    local_program = load_program("programs/local_quantum.iqoala")

    alice_inputs = [
        ProgramInput({"bob_id": bob_id, "state": i % 6}) for i in range(num_teleport)
    ]
    bob_inputs = [
        ProgramInput({"alice_id": alice_id, "state": i % 6})
        for i in range(num_teleport)
    ]
    local_inputs = [
        ProgramInput({"duration": local_busy_duration}) for _ in range(num_local)
    ]

    app_result = run_two_node_app_separate_inputs_plus_local_program(
        num_iterations=num_teleport,
        num_local_iterations=num_local,
        node1="alice",
        node2="bob",
        prog_node1=alice_program,
        prog_node1_inputs=alice_inputs,
        prog_node2=bob_program,
        prog_node2_inputs=bob_inputs,
        local_prog_node2=local_program,
        local_prog_node2_inputs=local_inputs,
        network_cfg=network_cfg,
        linear_for={"alice": False, "bob": False},
        linear_local=False,
    )

    alice_result = app_result.batch_results["alice"]
    bob_result = app_result.batch_results["bob"]
    local_result = app_result.batch_results["local"]
    # print(local_result)

    return TeleportResult(
        alice_result, bob_result, local_result, app_result.total_duration
    )


@dataclass
class DataPoint:
    num_teleport: int
    num_local: int
    tel_succ_prob: float
    tel_succ_prob_lower: float
    tel_succ_prob_upper: float
    loc_succ_prob: float
    loc_succ_prob_lower: float
    loc_succ_prob_upper: float
    makespan: float


@dataclass
class DataMeta:
    timestamp: str
    num_teleport: int
    num_local: int
    num_runs: int
    t1: float
    t2: float
    latency_factor: float
    cc_latency: float
    local_busy_duration: float
    network_bin_len: int
    network_period: int
    network_first_bin: int
    num_qubits_bob: int


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def wilson_score_interval(p_hat, n, z):
    denominator = 1 + z**2 / n
    centre_adjusted_probability = p_hat + z**2 / (2 * n)
    adjusted_standard_deviation = z * math.sqrt(
        (p_hat * (1 - p_hat) + z**2 / (4 * n)) / n
    )

    lower_bound = (
        centre_adjusted_probability - adjusted_standard_deviation
    ) / denominator
    upper_bound = (
        centre_adjusted_probability + adjusted_standard_deviation
    ) / denominator

    return (lower_bound, upper_bound)


def get_datapoint(
    num_runs: int,
    num_teleport: int,
    num_local: int,
    num_qubits_bob: int,
    t1: float,
    t2: float,
    cc_latency: float,
    network_bin_len: int,
    network_period: int,
    network_first_bin: int,
    local_busy_duration: int,
) -> DataPoint:
    teleport_successes: List[bool] = []
    local_successes: List[bool] = []
    makespans: List[float] = []

    for i in range(num_runs):
        print(f"run {i}/{num_runs}", end="\r")

        result = run_apps(
            num_teleport=num_teleport,
            num_local=num_local,
            num_qubits_bob=num_qubits_bob,
            t1=t1,
            t2=t2,
            cc_latency=cc_latency,
            network_period=network_period,
            network_bin_len=network_bin_len,
            network_first_bin=network_first_bin,
            local_busy_duration=local_busy_duration,
        )
        teleport_results = result.bob_result.results
        outcomes = [result.values["outcome"] for result in teleport_results]
        assert len(outcomes) == num_teleport
        teleport_successes.extend([outcomes[i] == 0 for i in range(num_teleport)])

        local_results = result.local_result.results
        outcomes = [result.values["outcome"] for result in local_results]
        assert len(outcomes) == num_local
        local_successes.extend([outcomes[i] == 1 for i in range(num_local)])

        makespans.append(result.total_duration)

    tel_avg_succ_prob = sum([s for s in teleport_successes if s]) / len(
        teleport_successes
    )
    tel_succ_prob_lower, tel_succ_prob_upper = wilson_score_interval(
        p_hat=tel_avg_succ_prob, n=len(teleport_successes), z=1.96
    )
    tel_lower_rounded = round(tel_succ_prob_lower, 3)
    tel_upper_rounded = round(tel_succ_prob_upper, 3)
    tel_succprob_rounded = round(tel_avg_succ_prob, 3)

    if len(local_successes) > 0:
        loc_avg_succ_prob = sum([s for s in local_successes if s]) / len(
            local_successes
        )
        loc_succ_prob_lower, loc_succ_prob_upper = wilson_score_interval(
            p_hat=loc_avg_succ_prob, n=len(local_successes), z=1.96
        )
        loc_lower_rounded = round(loc_succ_prob_lower, 3)
        loc_upper_rounded = round(loc_succ_prob_upper, 3)
        loc_succprob_rounded = round(loc_avg_succ_prob, 3)
    else:
        loc_avg_succ_prob = 0
        loc_succ_prob_lower = 0
        loc_succ_prob_upper = 0
        loc_lower_rounded = 0
        loc_upper_rounded = 0
        loc_succprob_rounded = 0

    makespan = sum(makespans) / len(makespans)

    print(
        f"teleport succ prob: {tel_succprob_rounded} ({tel_lower_rounded}, {tel_upper_rounded})"
    )
    print(
        f"local succ prob: {loc_succprob_rounded} ({loc_lower_rounded}, {loc_upper_rounded})"
    )
    print(f"makespan: {makespan:_}")

    return DataPoint(
        num_teleport=num_teleport,
        num_local=num_local,
        tel_succ_prob=tel_avg_succ_prob,
        tel_succ_prob_lower=tel_succ_prob_lower,
        tel_succ_prob_upper=tel_succ_prob_upper,
        loc_succ_prob=loc_avg_succ_prob,
        loc_succ_prob_lower=loc_succ_prob_lower,
        loc_succ_prob_upper=loc_succ_prob_upper,
        makespan=makespan,
    )


def run(
    output_dir: str, num_runs: int, num_teleport: int, num_local: int, dump: bool = True
):
    # LogManager.set_log_level("INFO")
    # LogManager.log_to_file("quantum_multitasking.log")
    # LogManager.set_task_log_level("INFO")
    # LogManager.set_task_log_level("WARNING")
    # LogManager.log_tasks_to_file("quantum_multitasking_tasks.log")

    start_time = time.time()
    timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")

    t1 = 1e10
    t2 = 1e7

    latency_factor = 0.01  # CC = 100_000
    cc_latency = latency_factor * t2
    local_busy_duration = 0

    network_bin_len = 100_000
    network_period = num_teleport * network_bin_len
    network_first_bin = 0

    num_qubits_bob = 10

    data_points: List[DataPoint] = []

    data_point = get_datapoint(
        num_runs=num_runs,
        num_teleport=num_teleport,
        num_local=num_local,
        num_qubits_bob=num_qubits_bob,
        t1=t1,
        t2=t2,
        cc_latency=cc_latency,
        network_bin_len=network_bin_len,
        network_period=network_period,
        network_first_bin=network_first_bin,
        local_busy_duration=local_busy_duration,
    )

    data_points.append(data_point)

    meta = DataMeta(
        timestamp=timestamp,
        num_teleport=num_teleport,
        num_local=num_local,
        num_runs=num_runs,
        t1=t1,
        t2=t2,
        latency_factor=latency_factor,
        cc_latency=cc_latency,
        local_busy_duration=local_busy_duration,
        network_bin_len=network_bin_len,
        network_period=network_period,
        network_first_bin=network_first_bin,
        num_qubits_bob=num_qubits_bob,
    )

    end_time = time.time()
    print(f"simulation took {end_time - start_time}s")

    if dump:
        data = Data(meta=meta, data_points=data_points)
        json_data = asdict(data)

        abs_dir = relative_to_cwd(f"data/{output_dir}")
        Path(abs_dir).mkdir(parents=True, exist_ok=True)
        last_path = os.path.join(abs_dir, "LAST.json")
        timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")
        with open(last_path, "w") as datafile:
            json.dump(json_data, datafile)
        with open(timestamp_path, "w") as datafile:
            json.dump(json_data, datafile)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--dump", "-d", action="store_true")
    parser.add_argument("--num_runs", "-n", type=int, required=True)
    parser.add_argument("--num_teleport", "-t", type=int, required=True)
    parser.add_argument("--num_local", "-l", type=int, required=True)

    args = parser.parse_args()
    dump = args.dump
    num_runs = args.num_runs
    num_teleport = args.num_teleport
    num_local = args.num_local

    run(
        f"sweep_teleport_local_{num_teleport}_{num_local}",
        num_runs=num_runs,
        num_teleport=num_teleport,
        num_local=num_local,
        dump=dump,
    )



from __future__ import annotations

import datetime
import json
import os
import time
from argparse import ArgumentParser
from dataclasses import asdict, dataclass
from pathlib import Path
from typing import Dict, List, Tuple

import netsquid as ns
from netqasm.lang.instr.flavour import NVFlavour, TrappedIonFlavour

from qoala.lang.ehi import UnitModule
from qoala.lang.parse import QoalaParser
from qoala.lang.program import QoalaProgram
from qoala.runtime.config import (
    LatenciesConfig,
    NtfConfig,
    ProcNodeConfig,
    ProcNodeNetworkConfig,
    TopologyConfig,
)
from qoala.runtime.program import BatchInfo, BatchResult, ProgramBatch, ProgramInput
from qoala.sim.build import build_network_from_config
from qoala.sim.network import ProcNodeNetwork
from qoala.util.runner import run_two_node_app


def relative_to_cwd(file: str) -> str:
    return os.path.join(os.path.dirname(__file__), file)


def create_procnode_cfg(
    name: str, id: int, num_qubits: int, hardware: str
) -> ProcNodeConfig:
    if hardware == "generic":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_config_uniform_default_params(num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("GenericNtf"),
        )
    elif hardware == "nv":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_nv_default_params(num_qubits=num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("NvNtf"),
        )
    elif hardware == "tri":
        return ProcNodeConfig(
            node_name=name,
            node_id=id,
            topology=TopologyConfig.perfect_tri_default_params(num_qubits=num_qubits),
            latencies=LatenciesConfig(qnos_instr_time=1000),
            ntf=NtfConfig.from_cls_name("TrappedIonNtf"),
        )


def get_client_config(id: int, hardware: str) -> ProcNodeConfig:
    # client only needs 1 qubit
    return create_procnode_cfg(
        name=f"client_{id}", id=id, num_qubits=1, hardware=hardware
    )


def get_server_config(id: int, num_qubits: int, hardware: str) -> ProcNodeConfig:
    return create_procnode_cfg(
        name="server", id=id, num_qubits=num_qubits, hardware=hardware
    )


def create_network(
    server_cfg: ProcNodeConfig,
    client_configs: List[ProcNodeConfig],
    num_clients: int,
) -> ProcNodeNetwork:
    assert len(client_configs) == num_clients

    node_cfgs = [server_cfg] + client_configs
    network_cfg = ProcNodeNetworkConfig.from_nodes_perfect_links(
        nodes=node_cfgs, link_duration=1000
    )
    return build_network_from_config(network_cfg)


@dataclass
class BqcResult:
    client_batches: List[Dict[int, ProgramBatch]]
    client_results: List[Dict[int, BatchResult]]


def load_server_program(remote_name: str, hardware: str) -> QoalaProgram:
    filename = f"vbqc_{hardware}_server.iqoala"
    path = os.path.join(os.path.dirname(__file__), filename)
    with open(path) as file:
        server_text = file.read()

    if hardware == "generic":
        flavour = None
    elif hardware == "nv":
        flavour = NVFlavour()
    elif hardware == "tri":
        flavour = TrappedIonFlavour()
    program = QoalaParser(server_text, flavour=flavour).parse()

    # Replace "client" by e.g. "client_1"
    program.meta.csockets[0] = remote_name
    program.meta.epr_sockets[0] = remote_name

    return program


def load_client_program(hardware: str) -> QoalaProgram:
    filename = f"vbqc_{hardware}_client.iqoala"
    path = os.path.join(os.path.dirname(__file__), filename)
    with open(path) as file:
        client_text = file.read()

    if hardware == "generic":
        flavour = None
    elif hardware == "nv":
        flavour = NVFlavour()
    elif hardware == "tri":
        flavour = TrappedIonFlavour()
    return QoalaParser(client_text, flavour=flavour).parse()


def create_server_batch(
    client_id: int,
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
    hardware: str,
) -> BatchInfo:
    server_program = load_server_program(
        remote_name=f"client_{client_id}", hardware=hardware
    )
    return BatchInfo(
        program=server_program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=0,
    )


def create_client_batch(
    inputs: List[ProgramInput],
    unit_module: UnitModule,
    num_iterations: int,
    hardware: str,
) -> BatchInfo:
    client_program = load_client_program(hardware=hardware)
    return BatchInfo(
        program=client_program,
        inputs=inputs,
        unit_module=unit_module,
        num_iterations=num_iterations,
        deadline=0,
    )


def run_bqc(
    hardware: str,
    alpha,
    beta,
    theta1,
    theta2,
    dummy0,
    dummy1,
    num_iterations: List[int],
    num_clients: int,
):
    ns.sim_reset()

    # server needs to have 2 qubits per client
    server_num_qubits = num_clients * 2
    server_config = get_server_config(
        id=0, num_qubits=server_num_qubits, hardware=hardware
    )
    client_configs = [get_client_config(i, hardware) for i in range(1, num_clients + 1)]

    network = create_network(server_config, client_configs, num_clients)
    server_procnode = network.nodes["server"]
    server_batches: Dict[int, ProgramBatch] = {}  # client ID -> server batch
    client_batches: Dict[int, ProgramBatch] = {}  # client ID -> client batch

    for client_id in range(1, num_clients + 1):
        # index in num_iterations list
        index = client_id - 1

        server_inputs = [
            ProgramInput({"client_id": client_id}) for _ in range(num_iterations[index])
        ]

        server_unit_module = UnitModule.from_full_ehi(server_procnode.memmgr.get_ehi())
        server_batch_info = create_server_batch(
            client_id=client_id,
            inputs=server_inputs,
            unit_module=server_unit_module,
            num_iterations=num_iterations[index],
            hardware=hardware,
        )

        server_batches[client_id] = server_procnode.submit_batch(server_batch_info)

    for client_id in range(1, num_clients + 1):
        # index in num_iterations list
        index = client_id - 1

        client_inputs = [
            ProgramInput(
                {
                    "server_id": 0,
                    "alpha": alpha,
                    "beta": beta,
                    "theta1": theta1,
                    "theta2": theta2,
                    "dummy0": dummy0,
                    "dummy1": dummy1,
                }
            )
            for _ in range(num_iterations[index])
        ]

        client_procnode = network.nodes[f"client_{client_id}"]

        client_unit_module = UnitModule.from_full_ehi(client_procnode.memmgr.get_ehi())
        client_batch_info = create_client_batch(
            client_inputs, client_unit_module, num_iterations[index], hardware
        )

        client_batches[client_id] = client_procnode.submit_batch(client_batch_info)
        batch_id = client_batches[client_id].batch_id
        server_pids = [inst.pid for inst in server_batches[client_id].instances]
        # print(
        #     f"client ID: {client_id}, batch ID: {batch_id}, server PIDs: {server_pids}"
        # )
        client_procnode.initialize_processes(
            remote_pids={batch_id: server_pids}, linear=True
        )

    client_pids = {
        server_batches[client_id].batch_id: [
            inst.pid for inst in client_batches[client_id].instances
        ]
        for client_id in range(1, num_clients + 1)
    }
    # print(f"client PIDs: {client_pids}")
    server_procnode.initialize_processes(remote_pids=client_pids, linear=True)

    network.start()
    start_time = ns.sim_time()
    ns.sim_run()
    end_time = ns.sim_time()
    makespan = end_time - start_time

    client_procnodes = [network.nodes[f"client_{i}"] for i in range(1, num_clients + 1)]
    client_batches = [node.get_batches() for node in client_procnodes]

    client_results: List[Dict[int, BatchResult]]
    client_results = [node.scheduler.get_batch_results() for node in client_procnodes]

    return BqcResult(client_batches, client_results), makespan


def check_computation(
    hardware: str,
    alpha,
    beta,
    theta1,
    theta2,
    dummy0,
    dummy1,
    expected,
    num_iterations,
    num_clients,
):
    ns.sim_reset()
    bqc_result, makespan = run_bqc(
        hardware=hardware,
        alpha=alpha,
        beta=beta,
        theta1=theta1,
        theta2=theta2,
        dummy0=dummy0,
        dummy1=dummy1,
        num_iterations=num_iterations,
        num_clients=num_clients,
    )

    batch_success_probabilities: List[float] = []

    for i in range(num_clients):
        assert len(bqc_result.client_results[i]) == 1
        batch_result = bqc_result.client_results[i][0]
        assert len(bqc_result.client_batches[i]) == 1
        program_batch = bqc_result.client_batches[i][0]
        batch_iterations = program_batch.info.num_iterations

        m2s = [result.values["m2"] for result in batch_result.results]
        correct_outcomes = len([m2 for m2 in m2s if m2 == expected])
        succ_prob = round(correct_outcomes / batch_iterations, 2)
        batch_success_probabilities.append(succ_prob)

    return batch_success_probabilities, makespan


def compute_succ_prob_computation(
    hardware: str,
    num_clients: int,
    num_iterations: List[int],
) -> Tuple[float, float]:
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    # theta1, theta2
    params = [
        (0, 0),
        (0, 7),
        (2, 22),
        (6, 15),
    ]

    all_probs = []
    all_makespans = []
    for theta1, theta2 in params:
        probs, makespan = check_computation(
            hardware=hardware,
            alpha=8,
            beta=24,
            theta1=theta1,
            theta2=theta2,
            dummy0=0,
            dummy1=0,
            expected=1,
            num_iterations=num_iterations,
            num_clients=num_clients,
        )
        all_probs.append(probs[0])
        all_makespans.append(makespan)

    prob = sum(all_probs) / len(all_probs)
    makespan = sum(all_makespans)
    return prob, makespan


def check_trap(
    hardware: str,
    alpha,
    beta,
    theta1,
    theta2,
    dummy0,
    dummy1,
    num_iterations,
    num_clients,
):
    ns.sim_reset()
    bqc_result, makespan = run_bqc(
        hardware=hardware,
        alpha=alpha,
        beta=beta,
        theta1=theta1,
        theta2=theta2,
        dummy0=dummy0,
        dummy1=dummy1,
        num_iterations=num_iterations,
        num_clients=num_clients,
    )

    batch_success_probabilities: List[float] = []

    for i in range(num_clients):
        assert len(bqc_result.client_results[i]) == 1
        batch_result = bqc_result.client_results[i][0]
        assert len(bqc_result.client_batches[i]) == 1
        program_batch = bqc_result.client_batches[i][0]
        batch_iterations = program_batch.info.num_iterations

        p1s = [result.values["p1"] for result in batch_result.results]
        p2s = [result.values["p2"] for result in batch_result.results]
        m1s = [result.values["m1"] for result in batch_result.results]
        m2s = [result.values["m2"] for result in batch_result.results]

        if dummy0 == 0:
            # corresponds to "dummy = 1"
            # do normal rotations on qubit 0
            # no rotations on qubit 1
            num_fails = len([(p, m) for (p, m) in zip(p1s, m2s) if p != m])
        else:  # dummy0 = 1
            # corresponds to "dummy = 2"
            # no rotations on qubit 0
            # do normal rotations on qubit 1
            num_fails = len([(p, m) for (p, m) in zip(p2s, m1s) if p != m])

        frac_fail = round(num_fails / batch_iterations, 2)
        batch_success_probabilities.append(1 - frac_fail)

    return batch_success_probabilities, makespan


def compute_succ_prob_trap(
    hardware: str,
    num_clients: int,
    num_iterations: List[int],
) -> Tuple[float, float]:
    ns.set_qstate_formalism(ns.qubits.qformalism.QFormalism.DM)

    # dummy0, dummy1
    params = [
        (0, 1),
        (1, 0),
    ]

    all_probs = []
    all_makespans = []
    for dummy0, dummy1 in params:
        probs, makespan = check_trap(
            hardware=hardware,
            alpha=8,
            beta=24,
            theta1=2,
            theta2=22,
            dummy0=dummy0,
            dummy1=dummy1,
            num_iterations=num_iterations,
            num_clients=num_clients,
        )
        all_probs.append(probs[0])
        all_makespans.append(makespan)

    prob = sum(all_probs) / len(all_probs)
    makespan = sum(all_makespans)
    return prob, makespan


def bqc_computation(hardware: str, num_clients: int, num_iterations: int):
    succ_probs, makespan = compute_succ_prob_computation(
        hardware=hardware,
        num_clients=num_clients,
        num_iterations=[num_iterations] * num_clients,
    )
    print(f"success probabilities: {succ_probs}")
    print(f"makespan: {makespan}")
    assert succ_probs == 1


def bqc_trap(hardware: str, num_clients: int, num_iterations: int):
    succ_probs, makespan = compute_succ_prob_trap(
        hardware=hardware,
        num_clients=num_clients,
        num_iterations=[num_iterations] * num_clients,
    )
    print(f"success probabilities: {succ_probs}")
    print(f"makespan: {makespan}")
    assert succ_probs == 1


@dataclass
class DataPoint:
    success: bool
    sim_duration: float


@dataclass
class DataMeta:
    timestamp: str
    num_clients: int
    num_iterations: int


@dataclass
class Data:
    meta: DataMeta
    data_points: List[DataPoint]


def run_hardware(hardware: str, num_clients: int, num_iterations: int) -> DataPoint:
    start = time.time()

    success: bool
    try:
        bqc_computation(hardware, num_clients, num_iterations)
        bqc_trap(hardware, num_clients, num_iterations)
        success = True
    except AssertionError:
        success = False

    end = time.time()
    duration = round(end - start, 2)
    print(f"duration: {duration} s")

    return DataPoint(success, duration)


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--num_clients", "-c", type=int, required=True)
    parser.add_argument("--num_iterations", "-n", type=int, required=True)

    args = parser.parse_args()
    num_iterations = args.num_iterations
    num_clients = args.num_clients

    timestamp = datetime.datetime.now().strftime("%Y%m%d_%H%M%S")

    # DEFAULT:
    # num_clients = 1
    # num_iterations = 100

    data_points: List[DataPoint] = []
    for hardware in ["generic", "nv", "tri"]:
        data_point = run_hardware(hardware, num_clients, num_iterations)
        data_points.append(data_point)

    abs_dir = relative_to_cwd(f"data")
    Path(abs_dir).mkdir(parents=True, exist_ok=True)
    last_path = os.path.join(abs_dir, "LAST.json")
    timestamp_path = os.path.join(abs_dir, f"{timestamp}.json")

    meta = DataMeta(
        timestamp=timestamp, num_clients=num_clients, num_iterations=num_iterations
    )
    data = Data(meta=meta, data_points=data_points)
    json_data = asdict(data)

    with open(last_path, "w") as datafile:
        json.dump(json_data, datafile)
    with open(timestamp_path, "w") as datafile:
        json.dump(json_data, datafile)



import os

import pytest
from netqasm.lang.instr.core import MeasInstruction, SetInstruction
from netqasm.lang.operand import Register, Template
from netqasm.lang.subroutine import Subroutine

from qoala.lang.hostlang import (
    AssignCValueOp,
    BasicBlockType,
    BusyOp,
    IqoalaSingleton,
    IqoalaTuple,
    IqoalaVector,
    IqoalaVectorElement,
    MultiplyConstantCValueOp,
    ReceiveCMsgOp,
    ReturnResultOp,
    RunRequestOp,
    RunSubroutineOp,
    SendCMsgOp,
)
from qoala.lang.parse import (
    HostCodeParser,
    IqoalaInstrParser,
    IqoalaMetaParser,
    LocalRoutineParser,
    QoalaParseError,
    QoalaParser,
    RequestRoutineParser,
)
from qoala.lang.program import LocalRoutine, ProgramMeta
from qoala.lang.request import (
    CallbackType,
    EprRole,
    EprType,
    QoalaRequest,
    RequestRoutine,
    RequestVirtIdMapping,
)
from qoala.lang.routine import RoutineMetadata
from qoala.util.tests import text_equal


def test_parse_incomplete_meta():
    text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
META_END
    """

    with pytest.raises(QoalaParseError):
        IqoalaMetaParser(text).parse()


def test_parse_meta_no_end():
    text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
    """

    with pytest.raises(QoalaParseError):
        IqoalaMetaParser(text).parse()


def test_parse_meta():
    text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    meta = IqoalaMetaParser(text).parse()

    assert meta == ProgramMeta(
        name="alice", parameters=[], csockets={0: "bob"}, epr_sockets={}
    )


def test_parse_meta_multiple_remotes():
    text = """
META_START
name: alice
parameters: theta1, theta2
csockets: 0 -> bob, 1 -> charlie
epr_sockets: 0 -> bob
META_END
    """

    meta = IqoalaMetaParser(text).parse()

    assert meta == ProgramMeta(
        name="alice",
        parameters=["theta1", "theta2"],
        csockets={0: "bob", 1: "charlie"},
        epr_sockets={0: "bob"},
    )


def test_parse_1_instr():
    text = """
x = assign_cval() : 1
    """

    instructions = IqoalaInstrParser(text).parse()

    assert len(instructions) == 1
    assert instructions[0] == AssignCValueOp(result=IqoalaSingleton("x"), value=1)


def test_parse_2_instr():
    text = """
x = assign_cval() : 1
y = assign_cval() : 17
    """

    instructions = IqoalaInstrParser(text).parse()

    assert len(instructions) == 2
    assert instructions[0] == AssignCValueOp(result=IqoalaSingleton("x"), value=1)
    assert instructions[1] == AssignCValueOp(result=IqoalaSingleton("y"), value=17)


def test_parse_busy():
    text = """
busy() : 250
    """

    instructions = IqoalaInstrParser(text).parse()

    assert len(instructions) == 1
    assert instructions[0] == BusyOp(value=250)


def test_parse_faulty_instr():
    text = """
x = assign_cval
    """

    with pytest.raises(QoalaParseError):
        IqoalaInstrParser(text).parse()


def test_parse_tuple():
    text = """
run_subroutine(tuple<x>) : subrt1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunSubroutineOp(
        result=None, values=IqoalaTuple(["x"]), subrt="subrt1"
    )


def test_parse_tuple_2_elements():
    text = """
run_subroutine(tuple<x; y>) : subrt1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunSubroutineOp(
        result=None, values=IqoalaTuple(["x", "y"]), subrt="subrt1"
    )


def test_parse_tuple_2_elements_and_return():
    text = """
tuple<m> = run_subroutine(tuple<x; y>) : subrt1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunSubroutineOp(
        result=IqoalaTuple(["m"]), values=IqoalaTuple(["x", "y"]), subrt="subrt1"
    )


def test_parse_tuple_2_elements_and_return_2_elements():
    text = """
tuple<m1; m2> = run_subroutine(tuple<x; y>) : subrt1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunSubroutineOp(
        result=IqoalaTuple(["m1", "m2"]),
        values=IqoalaTuple(["x", "y"]),
        subrt="subrt1",
    )


def test_parse_vector():
    text = """
my_vec<3> = run_subroutine() : subrt1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunSubroutineOp(
        result=IqoalaVector("my_vec", 3), values=IqoalaTuple([]), subrt="subrt1"
    )


def test_parse_vector_2():
    text = """
my_vec<3> = run_request() : req1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunRequestOp(
        result=IqoalaVector("my_vec", 3), values=IqoalaTuple([]), routine="req1"
    )


def test_parse_vector_with_var():
    text = """
my_vec<N> = run_request() : req1
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 1
    assert instructions[0] == RunRequestOp(
        result=IqoalaVector("my_vec", "N"), values=IqoalaTuple([]), routine="req1"
    )


def test_parse_block_header():
    text = "^b0 {type = CL}:"

    name, typ, duration = HostCodeParser("")._parse_block_header(text)
    assert name == "b0"
    assert typ == BasicBlockType.CL
    assert duration is None


def test_parse_block_header_with_deadlines():
    text = "^b0 {type = CL, deadlines = [b1: 1000]}:"

    name, typ, deadline = HostCodeParser("")._parse_block_header(text)
    assert name == "b0"
    assert typ == BasicBlockType.CL
    assert deadline == {"b1": 1000}


def test_parse_block():
    text = """
^b0 {type = CL}:
    x = assign_cval() : 1
    y = assign_cval() : 17
    """

    block = HostCodeParser("").parse_block(text)
    assert block.name == "b0"
    assert block.typ == BasicBlockType.CL
    assert len(block.instructions) == 2
    assert block.instructions[0] == AssignCValueOp(result=IqoalaSingleton("x"), value=1)
    assert block.instructions[1] == AssignCValueOp(
        result=IqoalaSingleton("y"), value=17
    )


def test_get_block_texts():
    text = """
^b0 {type = CL}:
    x = assign_cval() : 1
    y = assign_cval() : 17

^b1 {type = QL}:
    run_subroutine(tuple<x>) : subrt1
    """

    block_texts = HostCodeParser(text).get_block_texts()
    assert len(block_texts) == 2


def test_parse_multiple_blocks():
    text = """
^b0 {type = CL}:
    x = assign_cval() : 1
    y = assign_cval() : 17

^b1 {type = QL, deadlines = [b0: 2500]}:
    run_subroutine(tuple<x>) : subrt1
    """

    blocks = HostCodeParser(text).parse()
    assert len(blocks) == 2

    assert blocks[0].name == "b0"
    assert blocks[0].typ == BasicBlockType.CL
    assert len(blocks[0].instructions) == 2
    assert blocks[0].instructions[0] == AssignCValueOp(
        result=IqoalaSingleton("x"), value=1
    )
    assert blocks[0].instructions[1] == AssignCValueOp(
        result=IqoalaSingleton("y"), value=17
    )

    assert blocks[1].name == "b1"
    assert blocks[1].typ == BasicBlockType.QL
    assert blocks[1].deadlines == {"b0": 2500}
    assert len(blocks[1].instructions) == 1
    assert blocks[1].instructions[0] == RunSubroutineOp(
        result=None, values=IqoalaTuple(["x"]), subrt="subrt1"
    )


def test_parse_subrt():
    text = """
SUBROUTINE subrt1
    params: my_value
    returns: m
    uses: 
    keeps:
    request: 
  NETQASM_START
    set Q0 {my_value}
  NETQASM_END
    """

    parsed = LocalRoutineParser(text).parse()
    assert len(parsed) == 1
    assert "subrt1" in parsed
    subrt = parsed["subrt1"]

    expected_instrs = [
        SetInstruction(reg=Register.from_str("Q0"), imm=Template("my_value"))
    ]
    expected_args = ["my_value"]
    assert subrt == LocalRoutine(
        name="subrt1",
        subroutine=Subroutine(instructions=expected_instrs, arguments=expected_args),
        metadata=RoutineMetadata.use_none(),
        return_vars=["m"],
    )


def test_parse_subrt_2():
    text = """
SUBROUTINE my_subroutine
    params: param1, param2
    returns: result1, result2
    uses: 0, 1
    keeps: 
    request: 
  NETQASM_START
    set R0 {param1}
    set R1 {param2}
    meas Q0 M5
    meas Q1 M6
  NETQASM_END
    """

    parsed = LocalRoutineParser(text).parse()
    assert len(parsed) == 1
    assert "my_subroutine" in parsed
    subrt = parsed["my_subroutine"]

    expected_instrs = [
        SetInstruction(reg=Register.from_str("R0"), imm=Template("param1")),
        SetInstruction(reg=Register.from_str("R1"), imm=Template("param2")),
        MeasInstruction(reg0=Register.from_str("Q0"), reg1=Register.from_str("M5")),
        MeasInstruction(reg0=Register.from_str("Q1"), reg1=Register.from_str("M6")),
    ]
    expected_args = ["param1", "param2"]
    assert subrt == LocalRoutine(
        name="my_subroutine",
        subroutine=Subroutine(instructions=expected_instrs, arguments=expected_args),
        metadata=RoutineMetadata.free_all([0, 1]),
        return_vars=["result1", "result2"],
    )


def test_parse_subrt_3():
    text = """
SUBROUTINE my_subroutine
    params: param1, param2
    returns: outcomes<10>
    uses: 0, 1
    keeps: 
    request: 
  NETQASM_START
    set R0 {param1}
    set R1 {param2}
    meas Q0 M5
    meas Q1 M6
  NETQASM_END
    """

    parsed = LocalRoutineParser(text).parse()
    assert len(parsed) == 1
    assert "my_subroutine" in parsed
    subrt = parsed["my_subroutine"]

    expected_instrs = [
        SetInstruction(reg=Register.from_str("R0"), imm=Template("param1")),
        SetInstruction(reg=Register.from_str("R1"), imm=Template("param2")),
        MeasInstruction(reg0=Register.from_str("Q0"), reg1=Register.from_str("M5")),
        MeasInstruction(reg0=Register.from_str("Q1"), reg1=Register.from_str("M6")),
    ]
    expected_args = ["param1", "param2"]
    assert subrt == LocalRoutine(
        name="my_subroutine",
        subroutine=Subroutine(instructions=expected_instrs, arguments=expected_args),
        metadata=RoutineMetadata.free_all([0, 1]),
        return_vars=[IqoalaVector("outcomes", 10)],
    )


def test_parse_multiple_subrt():
    text = """
SUBROUTINE subrt1
    params: param1
    returns: m
    uses: 0
    keeps:
    request: 
  NETQASM_START
    set R0 {param1}
    meas Q0 M0
  NETQASM_END

SUBROUTINE subrt2
    params: theta
    returns: 
    uses: 
    keeps:
    request: 
  NETQASM_START
    set R0 {theta}
  NETQASM_END
    """

    parsed = LocalRoutineParser(text).parse()
    assert len(parsed) == 2
    assert "subrt1" in parsed
    assert "subrt2" in parsed
    subrt1 = parsed["subrt1"]
    subrt2 = parsed["subrt2"]

    expected_instrs_1 = [
        SetInstruction(reg=Register.from_str("R0"), imm=Template("param1")),
        MeasInstruction(reg0=Register.from_str("Q0"), reg1=Register.from_str("M0")),
    ]
    expected_args_1 = ["param1"]
    expected_instrs_2 = [
        SetInstruction(reg=Register.from_str("R0"), imm=Template("theta")),
    ]
    expected_args_2 = ["theta"]

    assert subrt1 == LocalRoutine(
        name="subrt1",
        subroutine=Subroutine(
            instructions=expected_instrs_1, arguments=expected_args_1
        ),
        metadata=RoutineMetadata.free_all([0]),
        return_vars=["m"],
    )
    assert subrt2 == LocalRoutine(
        name="subrt2",
        subroutine=Subroutine(
            instructions=expected_instrs_2, arguments=expected_args_2
        ),
        metadata=RoutineMetadata.use_none(),
        return_vars=[],
    )


def test_parse_invalid_subrt():
    text = """
SUBROUTINE my_subroutine
    params: param1, param2
    returns: result1, result2
    request: 
  NETQASM_START
    set R0 {param3}
  NETQASM_END
    """

    with pytest.raises(QoalaParseError):
        LocalRoutineParser(text).parse()


def test_parse_request():
    text = """
REQUEST req1
  callback_type: wait_all
  callback: 
  return_vars: 
  remote_id: 1
  epr_socket_id: 0
  num_pairs: 5
  virt_ids: all 0
  timeout: 1000
  fidelity: 0.65
  typ: create_keep
  role: create
    """

    parsed = RequestRoutineParser(text).parse()
    assert len(parsed) == 1
    assert "req1" in parsed
    routine = parsed["req1"]

    assert routine == RequestRoutine(
        name="req1",
        return_vars=[],
        callback_type=CallbackType.WAIT_ALL,
        callback=None,
        request=QoalaRequest(
            name="req1",
            remote_id=1,
            epr_socket_id=0,
            num_pairs=5,
            virt_ids=RequestVirtIdMapping.from_str("all 0"),
            timeout=1000,
            fidelity=0.65,
            typ=EprType.CREATE_KEEP,
            role=EprRole.CREATE,
        ),
    )


def test_parse_request_2():
    text = """
REQUEST req1
  callback_type: sequential
  callback: subrt1
  return_vars: outcomes<3>
  remote_id: 1
  epr_socket_id: 0
  num_pairs: 3
  virt_ids: custom 1, 2, 3
  timeout: 1000
  fidelity: 0.65
  typ: measure_directly
  role: receive
    """

    parsed = RequestRoutineParser(text).parse()
    assert len(parsed) == 1
    assert "req1" in parsed
    routine = parsed["req1"]

    assert routine == RequestRoutine(
        name="req1",
        return_vars=[IqoalaVector("outcomes", 3)],
        callback_type=CallbackType.SEQUENTIAL,
        callback="subrt1",
        request=QoalaRequest(
            name="req1",
            remote_id=1,
            epr_socket_id=0,
            num_pairs=3,
            virt_ids=RequestVirtIdMapping.from_str("custom 1, 2, 3"),
            timeout=1000,
            fidelity=0.65,
            typ=EprType.MEASURE_DIRECTLY,
            role=EprRole.RECEIVE,
        ),
    )


def test_parse_request_with_template():
    text = """
REQUEST req1
  callback_type: wait_all
  callback: 
  return_vars: 
  remote_id: {client_id}
  epr_socket_id: 0
  num_pairs: 3
  virt_ids: custom 1, 2, 3
  timeout: 1000
  fidelity: 0.65
  typ: measure_directly
  role: receive
    """

    parsed = RequestRoutineParser(text).parse()
    assert len(parsed) == 1
    assert "req1" in parsed
    routine = parsed["req1"]

    assert routine == RequestRoutine(
        name="req1",
        return_vars=[],
        callback_type=CallbackType.WAIT_ALL,
        callback=None,
        request=QoalaRequest(
            name="req1",
            remote_id=Template("client_id"),
            epr_socket_id=0,
            num_pairs=3,
            virt_ids=RequestVirtIdMapping.from_str("custom 1, 2, 3"),
            timeout=1000,
            fidelity=0.65,
            typ=EprType.MEASURE_DIRECTLY,
            role=EprRole.RECEIVE,
        ),
    )


def test_parse_request_with_vector_template():
    text = """
REQUEST req1
  callback_type: wait_all
  callback: 
  return_vars: outcomes<{N}>
  remote_id: 0
  epr_socket_id: 0
  num_pairs: 3
  virt_ids: custom 1, 2, 3
  timeout: 1000
  fidelity: 0.65
  typ: measure_directly
  role: receive
    """

    parsed = RequestRoutineParser(text).parse()
    assert len(parsed) == 1
    assert "req1" in parsed
    routine = parsed["req1"]

    assert routine == RequestRoutine(
        name="req1",
        return_vars=[IqoalaVector("outcomes", Template("N"))],
        callback_type=CallbackType.WAIT_ALL,
        callback=None,
        request=QoalaRequest(
            name="req1",
            remote_id=0,
            epr_socket_id=0,
            num_pairs=3,
            virt_ids=RequestVirtIdMapping.from_str("custom 1, 2, 3"),
            timeout=1000,
            fidelity=0.65,
            typ=EprType.MEASURE_DIRECTLY,
            role=EprRole.RECEIVE,
        ),
    )


def test_parse_multiple_request():
    text = """
REQUEST req1
  callback_type: wait_all
  callback: 
  return_vars: 
  remote_id: 1
  epr_socket_id: 0
  num_pairs: 5
  virt_ids: all 0
  timeout: 1000
  fidelity: 0.65
  typ: create_keep
  role: create

REQUEST req2
  callback_type: sequential
  callback: subrt1
  return_vars: 
  remote_id: 1
  epr_socket_id: 0
  num_pairs: 3
  virt_ids: increment 1
  timeout: 1000
  fidelity: 0.65
  typ: measure_directly
  role: receive
    """

    parsed = RequestRoutineParser(text).parse()
    assert len(parsed) == 2
    assert "req1" in parsed
    assert "req2" in parsed
    routine1 = parsed["req1"]
    routine2 = parsed["req2"]

    assert routine1 == RequestRoutine(
        name="req1",
        return_vars=[],
        callback_type=CallbackType.WAIT_ALL,
        callback=None,
        request=QoalaRequest(
            name="req1",
            remote_id=1,
            epr_socket_id=0,
            num_pairs=5,
            virt_ids=RequestVirtIdMapping.from_str("all 0"),
            timeout=1000,
            fidelity=0.65,
            typ=EprType.CREATE_KEEP,
            role=EprRole.CREATE,
        ),
    )
    assert routine2 == RequestRoutine(
        name="req2",
        return_vars=[],
        callback_type=CallbackType.SEQUENTIAL,
        callback="subrt1",
        request=QoalaRequest(
            name="req2",
            remote_id=1,
            epr_socket_id=0,
            num_pairs=3,
            virt_ids=RequestVirtIdMapping.from_str("increment 1"),
            timeout=1000,
            fidelity=0.65,
            typ=EprType.MEASURE_DIRECTLY,
            role=EprRole.RECEIVE,
        ),
    )


def test_parse_invalid_request():
    text = """
REQUEST req1
  callback_type: wait_all
  callback: 
  return_vars: 
  remote_id: 1
  epr_socket_id: 0
  num_pairs: 3
  virt_ids: all 0
  timeout: 1000
  fidelity: 0.65
  typ: invalid
  role: receive
    """

    # invalid 'typ' value

    with pytest.raises(QoalaParseError):
        RequestRoutineParser(text).parse()


DEFAULT_META = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
"""


def test_parse_program():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    program_text = """
^b0 {type = CL}:
    my_value = assign_cval() : 1
    remote_id = assign_cval() : 0
    send_cmsg(remote_id, my_value)
    received_value = recv_cmsg(remote_id)
    new_value = assign_cval() : 3
    my_value = add_cval_c(new_value, new_value)

^b1 {type = QL}:
    tuple<m> = run_subroutine(tuple<my_value>) : subrt1

^b2 {type = QC}:
    tuple<m> = run_request(tuple<my_value>) : req1

^b3 {type = CL}:
    return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: my_value
    returns: m
    uses: 0
    keeps:
    request: 
  NETQASM_START
    set Q0 0
    rot_z Q0 {my_value} 4
    meas Q0 M0
    ret_reg M0
  NETQASM_END
    """

    req_text = """
REQUEST req1
  callback_type: wait_all
  callback: 
  return_vars: 
  remote_id: 1
  epr_socket_id: 0
  num_pairs: 3
  virt_ids: increment 1
  timeout: 1000
  fidelity: 0.65
  typ: measure_directly
  role: receive
    """

    parsed_program = QoalaParser(
        meta_text=meta_text,
        host_text=program_text,
        subrt_text=subrt_text,
        req_text=req_text,
    ).parse()
    assert len(parsed_program.blocks) == 4
    assert len(parsed_program.instructions) == 9
    assert "subrt1" in parsed_program.local_routines
    assert len(parsed_program.request_routines) == 1
    assert "req1" in parsed_program.request_routines


def test_parse_program_no_requests():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    program_text = """
^b0 {type = CL}:
    my_value = assign_cval() : 1
    remote_id = assign_cval() : 0
    send_cmsg(remote_id, my_value)
    received_value = recv_cmsg(remote_id)
    new_value = assign_cval() : 3
    my_value = add_cval_c(new_value, new_value)

^b1 {type = QL}:
    tuple<m> = run_subroutine(tuple<my_value>) : subrt1

^b1 {type = CL}:
    return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: my_value
    returns: m
    uses: 0
    keeps:
    request: 
  NETQASM_START
    set Q0 0
    rot_z Q0 {my_value} 4
    meas Q0 M0
    ret_reg M0
  NETQASM_END
    """

    req_text = """
    """

    parsed_program = QoalaParser(
        meta_text=meta_text,
        host_text=program_text,
        subrt_text=subrt_text,
        req_text=req_text,
    ).parse()
    assert len(parsed_program.blocks) == 3
    assert len(parsed_program.instructions) == 8
    assert "subrt1" in parsed_program.local_routines
    assert len(parsed_program.request_routines) == 0


def test_parse_program_no_subroutines():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    program_text = """
^b0 {type = CL}:
    my_value = assign_cval() : 1
    remote_id = assign_cval() : 0
    send_cmsg(remote_id, my_value)
    received_value = recv_cmsg(remote_id)
    new_value = assign_cval() : 3
    m = add_cval_c(new_value, new_value)
    return_result(m)
    """

    subrt_text = """
    """

    req_text = """
    """

    parsed_program = QoalaParser(
        meta_text=meta_text,
        host_text=program_text,
        subrt_text=subrt_text,
        req_text=req_text,
    ).parse()
    assert len(parsed_program.blocks) == 1
    assert len(parsed_program.instructions) == 7
    assert len(parsed_program.request_routines) == 0


def test_parse_program_invalid_subrt_reference():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    program_text = """
^b0 {type = CL}:
    my_value = assign_cval() : 1
^b1 {type = QL}:
    tuple<m> = run_subroutine(tuple<my_value>) : non_existing_subrt
^b2 {type = CL}:
    return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: my_value
    returns: m
    uses: 
    keeps:
    request: 
  NETQASM_START
    set R0 0
  NETQASM_END
    """

    with pytest.raises(QoalaParseError):
        QoalaParser(
            meta_text=meta_text,
            host_text=program_text,
            subrt_text=subrt_text,
            req_text="",
        ).parse()


def test_parse_program_invalid_req_routine_reference():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    program_text = """
^b0 {type = CL}:
    my_value = assign_cval() : 1
^b1 {type = QL}:
    tuple<m> = run_request(tuple<my_value>) : non_existing_req
^b2 {type = CL}:
    return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: my_value
    returns: m
    uses: 
    keeps:
    request: 
  NETQASM_START
    set R0 0
  NETQASM_END
    """

    with pytest.raises(QoalaParseError):
        QoalaParser(
            meta_text=meta_text,
            host_text=program_text,
            subrt_text=subrt_text,
            req_text="",
        ).parse()


def test_split_text():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    host_text = """
^b0 {type = CL}:
    m = assign_cval() : 1
    return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: 
    returns: 
    request: 
  NETQASM_START
    set Q0 0
  NETQASM_END
    """

    req_text = """
REQUEST req1
  callback_type: wait_all
  callback:
  return_vars: 
  remote_id: {client_id}
  epr_socket_id: 0
  num_pairs: 1
  virt_ids: all 0
  timeout:1000
  fidelity: 1.0
  typ: create_keep
  role: receive
    """

    text = meta_text + host_text + subrt_text + req_text
    parser = QoalaParser(text)

    assert text_equal(parser._meta_text, meta_text)
    assert text_equal(parser._host_text, host_text)
    assert text_equal(parser._subrt_text, subrt_text)
    assert text_equal(parser._req_text, req_text)


def test_split_text_multiple_subroutines():
    meta_text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END
    """

    host_text = """
^b0 {type = CL}:
    m = assign_cval() : 1
    return_result(m)
    """

    subrt_text = """
SUBROUTINE subrt1
    params: 
    returns: 
    uses:
    keeps:
    request: 
  NETQASM_START
    set Q0 0
  NETQASM_END

SUBROUTINE subrt2
    params: 
    returns: 
    uses:
    keeps:
    request: 
  NETQASM_START
    set Q7 7
  NETQASM_END
    """

    text = meta_text + host_text + subrt_text
    parser = QoalaParser(text)

    assert text_equal(parser._meta_text, meta_text)
    assert text_equal(parser._host_text, host_text)
    assert text_equal(parser._subrt_text, subrt_text)


def test_split_text_no_subroutines():
    meta_text = """
META_START
    name: server
    parameters: client_id
    csockets: 0 -> client
    epr_sockets: 0 -> client
META_END
    """

    host_text = """
    ^b0 {type = CL}:
        remote_id = assign_cval() : {client_id}
    ^b1 {type = QC}:
        run_request() : req1
        """

    req_text = """
REQUEST req1
    callback_type: wait_all
    callback:
  return_vars: 
    remote_id: {client_id}
    epr_socket_id: 0
    num_pairs: 1
    virt_ids: all 0
    timeout:1000
    fidelity: 1.0
    typ: create_keep
    role: receive
        """

    text = meta_text + host_text + req_text
    parser = QoalaParser(text)

    assert text_equal(parser._meta_text, meta_text)
    assert text_equal(parser._host_text, host_text)
    assert text_equal(parser._req_text, req_text)


def test_parse_program_single_text():
    text = """
META_START
name: alice
parameters: 
csockets: 0 -> bob
epr_sockets: 
META_END

^b0 {type = CL}:
    my_value = assign_cval() : 1
    remote_id = assign_cval() : 0
    send_cmsg(remote_id, my_value)
    received_value = recv_cmsg(remote_id)
    new_value = assign_cval() : 3
    my_value = add_cval_c(new_value, new_value)
^b1 {type = QL}:
    tuple<m> = run_subroutine(tuple<my_value>) : subrt1
^b2 {type = CL}:
    return_result(m)

SUBROUTINE subrt1
    params: my_value
    returns: m
    uses: 0
    keeps:
    request: 
  NETQASM_START
    set Q0 0
    rot_z Q0 {my_value} 4
    meas Q0 M0
    ret_reg M0
  NETQASM_END
    """

    parsed_program = QoalaParser(text).parse()
    assert len(parsed_program.blocks) == 3
    assert len(parsed_program.instructions) == 8
    assert "subrt1" in parsed_program.local_routines


def test_parse_file():
    path = os.path.join(os.path.dirname(__file__), "test_parse_server.iqoala")
    with open(path) as file:
        text = file.read()
    parsed_program = QoalaParser(text).parse()
    assert len(parsed_program.blocks) == 9
    assert len(parsed_program.instructions) == 11
    assert "create_epr_0" in parsed_program.local_routines
    assert "create_epr_1" in parsed_program.local_routines
    assert "local_cphase" in parsed_program.local_routines
    assert "meas_qubit_0" in parsed_program.local_routines
    assert "meas_qubit_1" in parsed_program.local_routines
    assert "req0" in parsed_program.request_routines
    assert "req1" in parsed_program.request_routines
    assert parsed_program.local_routines["create_epr_0"].request_name == "req0"
    assert parsed_program.local_routines["create_epr_1"].request_name == "req1"


def test_parse_file_2():
    path = os.path.join(os.path.dirname(__file__), "test_parse_client.iqoala")
    with open(path) as file:
        text = file.read()
    parsed_program = QoalaParser(text).parse()
    assert len(parsed_program.blocks) == 6
    assert len(parsed_program.instructions) == 19
    assert "create_epr_0" in parsed_program.local_routines
    assert "post_epr_0" in parsed_program.local_routines
    assert "create_epr_1" in parsed_program.local_routines
    assert "post_epr_1" in parsed_program.local_routines
    assert "req0" in parsed_program.request_routines
    assert "req1" in parsed_program.request_routines
    assert parsed_program.local_routines["create_epr_0"].request_name == "req0"
    assert parsed_program.local_routines["create_epr_1"].request_name == "req1"


def test_vector_indexing():
    text = """
    a<3> = run_request() : req1    
    send_cmsg(a[0], b)
    b = recv_cmsg(a[0])
    b = mult_const(a[0]) : 1
    """

    instructions = IqoalaInstrParser(text).parse()

    assert len(instructions) == 4
    csocket = IqoalaVectorElement("a", 0)
    remote_node = IqoalaSingleton("b")
    assert instructions[0] == RunRequestOp(
        result=IqoalaVector("a", 3), values=IqoalaTuple([]), routine="req1"
    )
    assert instructions[1] == SendCMsgOp(csocket, remote_node)
    assert instructions[2] == ReceiveCMsgOp(csocket, remote_node)
    assert instructions[3] == MultiplyConstantCValueOp(remote_node, csocket, 1)


def test_vector_indexing_error():
    text = """
    send_cmsg(a[0], b)
    b = recv_cmsg(a[0])
    b = mult_const(a[0]) : 1
    """

    with pytest.raises(QoalaParseError):
        IqoalaInstrParser(text).parse()


def test_arg_vector_passing():
    text = """
    a<3> = run_request() : req1    
    return_result(a)
    """

    instructions = IqoalaInstrParser(text).parse()
    assert len(instructions) == 2
    vec = IqoalaVector("a", 3)
    assert instructions[0] == RunRequestOp(
        result=vec, values=IqoalaTuple([]), routine="req1"
    )
    assert instructions[1] == ReturnResultOp(vec)


if __name__ == "__main__":
    test_parse_incomplete_meta()
    test_parse_meta_no_end()
    test_parse_meta()
    test_parse_meta_multiple_remotes()
    test_parse_1_instr()
    test_parse_2_instr()
    test_parse_busy()
    test_parse_faulty_instr()
    test_parse_tuple()
    test_parse_tuple_2_elements()
    test_parse_tuple_2_elements_and_return()
    test_parse_tuple_2_elements_and_return_2_elements()
    test_parse_vector()
    test_parse_vector_2()
    test_parse_vector_with_var()
    test_parse_block_header()
    test_parse_block_header_with_deadlines()
    test_parse_block()
    test_get_block_texts()
    test_parse_multiple_blocks()
    test_parse_subrt()
    test_parse_subrt_2()
    test_parse_subrt_3()
    test_parse_multiple_subrt()
    test_parse_invalid_subrt()
    test_parse_request()
    test_parse_request_2()
    test_parse_request_with_template()
    test_parse_request_with_vector_template()
    test_parse_multiple_request()
    test_parse_invalid_request()
    test_parse_program()
    test_parse_program_no_requests()
    test_parse_program_no_subroutines()
    test_parse_program_invalid_subrt_reference()
    test_parse_program_invalid_req_routine_reference()
    test_split_text()
    test_split_text_multiple_subroutines()
    test_split_text_no_subroutines()
    test_parse_program_single_text()
    test_parse_file()
    test_parse_file_2()
    test_vector_indexing()
    test_vector_indexing_error()
    test_arg_vector_passing()
